import numpy as np
import scipy.optimize as opt
from scipy import interpolate


def _error_func(data, a, b, c, d):
    """
    second order polynomial + exponential function.
    """
    return a * data + b * data**2 + c ** (data + d)


def _reject_outliers(data: np.ndarray, sigma: int = 2):
    """
    Filter data outside ``sigma`` standard deviations
    """
    return data[abs(data - np.median(data)) < (sigma * np.std(data))]


def _running_stddev(x, y, boxsize: int = 30, sigma: int = 2):
    """
    Calculate the standard deviations in a moving window with size ``boxsize``.
    Data outside ``sigma`` standard deviations are filtered.
    """
    # remove nan's and inf's
    idx = ~(np.isinf(y) | np.isnan(y))
    ydata = y[idx].copy()
    xvals = x[idx].copy()

    fr = boxsize // 2
    to = len(xvals) - fr

    xvals = xvals[fr:-fr]

    stddev = np.array(
        [
            np.std(_reject_outliers(ydata[j - fr : j + fr - 1], sigma=sigma))
            for j in np.arange(fr, to, 1)
        ]
    )

    return xvals, stddev


def calculate_error_parameters(y_true, y_model, boxsize=30, sigma=2, func=_error_func):
    """
    Fit function ``func`` to the running standard deviation of a dataset.
    """
    xvals, stddev = _running_stddev(y_true, y_model, boxsize=boxsize, sigma=sigma)
    f_results = opt.curve_fit(f=func, xdata=xvals, ydata=stddev)
    error_params = f_results[0]

    xlimits = xvals.min(), xvals.max()

    return error_params, xlimits


def calculate_spline_parameters(y_true, y_model, boxsize=30, sigma=2, s=3, fix0=True):

    xvals, stddev = _running_stddev(y_true, y_model, boxsize=boxsize, sigma=sigma)
    weights = np.onex(len(xvals))
    if fix0:
        weights[0] = 10
    params = interpolate.splrep(xvals, stddev, w=weights, s=s)

    xlimits = xvals.min(), xvals.max()

    return params, xlimits



from typing import Protocol

import pandas as pd

from MagmaPandas.enums import Datatype, Unit


class Magma(Protocol):
    _units: Unit
    _datatype: Datatype
    _weights: pd.Series

    @property
    def _no_data(self): ...

    @property
    def _total(self): ...

    @property
    def units(self): ...

    @property
    def weights(self): ...

    @property
    def elements(self): ...

    def moles(self): ...

    @property
    def wt_pc(self): ...

    def cations(self): ...

    @property
    def oxygen(self): ...

    def convert_ppm_wtPercent(self): ...

    def _convert_moles_wtPercent(self): ...

    def recalculate(self): ...

    def normalise(self): ...

    def random_sample(self): ...



"""
===========
MagmaFrames
===========
Module with the generic MagmaFrame class.
"""

import re
from typing import Dict, List

import elementMass as e
import numpy as np
import pandas as pd
from MagmaPandas.Elements import element_weights, oxide_compositions
from MagmaPandas.enums import Datatype, Unit
from MagmaPandas.MagmaFrames.protocols import Magma
from MagmaPandas.parse_io.validate import _check_argument, _check_attribute
from typing_extensions import Self


class MagmaFrame(pd.DataFrame):
    """
    Generic MagmaPandas DataFrame class for geochemical data.

    Parameters
    ----------
    data : ndarray (structured or homogeneous), Iterable, dict, or DataFrame
        geochemical data with elements or oxides in columns
    units : None, str
        data units, either "mol fraction", "wt. %" or "ppm"
    datatype : None, str
        datatype either "cation" or "oxide"
    weights : None, pandas Series
        atomic weights of elements or oxides in the MagmaFrame
    """

    # New attributes
    _metadata = ["_weights", "_units", "_datatype"]

    @_check_argument("units", [None, "mol fraction", "wt. %", "ppm"])
    @_check_argument("datatype", [None, "cation", "oxide"])
    def __init__(
        self,
        data=None,
        *args,
        units: None | str = None,
        datatype: None | str = None,
        weights: pd.Series = None,
        **kwargs,
    ) -> None:
        self._units: Unit = Unit(units)
        self._datatype: Datatype = Datatype(datatype)

        super().__init__(data, **kwargs)

        if not self._total:
            total_regex = re.search(
                "total", "".join(map(str, self.columns.to_list())), re.IGNORECASE
            )
            if total_regex:
                self.rename(columns={total_regex[0]: "total"}, inplace=True)

        if weights is not None:
            self._weights = weights.copy()
        elif not hasattr(self, "_weights"):
            self._weights = element_weights.weights_as_series(self.columns)

    @property
    def _constructor(self):
        """This is the key to letting Pandas know how to keep
        derivatives of `MagmaBase` the same type as yours.  It should
        be enough to return the name of the Class.  However, in
        some cases, `__finalize__` is not called and `new attributes` are
        not carried over.  We can fix that by constructing a callable
        that makes sure to call `__finalize__` every time."""

        def _c(*args, **kwargs):
            if (weights := getattr(self, "_weights", None)) is not None:
                weights = weights.copy(deep=True)

            current_class = type(self)

            return current_class(*args, weights=weights, **kwargs).__finalize__(self)

        return _c

    @property
    def _constructor_sliced(self):
        from MagmaPandas.MagmaSeries import MagmaSeries

        def _c(*args, **kwargs):
            if (weights := getattr(self, "_weights", None)) is not None:
                weights = weights.copy(deep=True)

            return MagmaSeries(*args, weights=weights, **kwargs).__finalize__(self)

        return _c

    @property
    def _no_data(self) -> List:
        """
        Names of all columns without chemical data
        """
        no_data = list(self.columns.difference(self.elements))
        if "total" in no_data:
            no_data.remove("total")
        return no_data

    @property
    def _total(self) -> bool:
        """
        Dataframe contains column with totals
        """

        return "total" in self.columns

    @property
    def units(self) -> str:
        """
        Datatype and units.
        """
        return f"{self._datatype.value} {self._units.value}"

    @units.setter
    def units(self, value):
        print("units are read only")

    @property
    def weights(self) -> pd.Series:
        """
        Atomic weights of all elements in the MagmaFrame.
        """
        return self._weights.copy()

    @property
    def elements(self) -> List[str]:
        """
        Names of all elements in the MagmaFrame.
        """
        return list(self._weights.index).copy()

    # @_check_attribute("_units", ["wt. %", "ppm"])
    def moles(self, normalise=True) -> Self:
        """
        Data converted to mol fraction.
        """

        if self._units == Unit.MOL_FRACTIONS:
            return self.copy()

        if self._units == Unit.WT_PERCENT:
            return self._convert_moles_wtPercent(normalise=normalise)
        elif self._units == Unit.PPM:
            return self.convert_ppm_wtPercent()._convert_moles_wtPercent(
                normalise=normalise
            )

        return self.copy()

    def wt_pc(self, normalise=True) -> Self:
        """
        Data converted to wt. %.
        """

        if self._units == Unit.WT_PERCENT:
            return self.copy()

        if self._units == Unit.MOL_FRACTIONS:
            return self._convert_moles_wtPercent(normalise=normalise)
        elif self._units == Unit.PPM:
            return self.convert_ppm_wtPercent()

        return self.copy()

    @property
    def ppm(self) -> Self:
        """
        Data converted to ppm.
        """

        if self._units == Unit.PPM:
            return self.copy()

        if self._units == Unit.WT_PERCENT:
            return self.convert_ppm_wtPercent()
        elif self._units == Unit.MOL_FRACTIONS:
            return self._convert_moles_wtPercent().convert_ppm_wtPercent()

        return self.copy()

    def cations(self, normalise=True, norm_to=1, mol_fractions=True) -> Self:
        """
        Data converted to cation mol fraction
        """
        # Calculate oxide moles
        if (self._datatype == Datatype.CATION) & (
            mol_fractions & (self._units == Unit.MOL_FRACTIONS)
        ):
            return self.copy()

        moles = self[self.elements].moles(normalise=False)

        # Calculate cation moles
        cations_per_oxide = oxide_compositions.cation_amount(moles.elements)
        cations = moles[moles.elements].mul(cations_per_oxide)
        # Rename columns to cations
        cations.columns = oxide_compositions.cation_names(moles.elements)

        cations._datatype = Datatype.CATION
        cations = cations.recalculate()

        if not mol_fractions:
            cations = cations[cations.elements].mul(cations.weights)
            cations._units = Unit.WT_PERCENT
            norm_to = 100

        if not normalise:
            cations["total"] = cations.sum(axis=1)
            return cations

        # Normalise to 1
        total = cations.sum(axis=1)
        cations = cations.div(total, axis=0) * norm_to
        cations["total"] = norm_to

        # Set the right datatype and elements
        cations._datatype = Datatype.CATION
        cations.recalculate(inplace=True)

        return cations

    def oxides(self, normalise=True, oxidation_state: Dict[str, int] = {}) -> Self:
        """
        Data converted to oxides
        """

        if (self._datatype == Datatype.OXIDE) & (not bool(oxidation_state)):
            return self.copy()

        units = self._units

        cations = self[self.elements].cations(
            normalise=False
        )  # convert to cation mol fractions
        cation_names = cations.elements
        cation_element_names = [
            re.sub(r"\d+", "", e) for e in cation_names
        ]  # strip numbers/charges from the names

        cation_names_new = [
            (
                i
                if oxidation_state.get(j, None) is None
                else f"{j}{int(oxidation_state[j])}"
            )
            for i, j in zip(cation_names, cation_element_names)
        ]  # new names include oxidation state for elements with non-default values

        oxide_names = e.get_oxide_names(cation_names_new)
        cations_per_oxide = e.cation_numbers(oxide_names)

        oxides = cations.rename(
            columns={cation: oxide for cation, oxide in zip(cation_names, oxide_names)}
        ).recalculate()  # rename to oxides

        oxides = oxides.div(cations_per_oxide, axis=1)  # recalculate to oxides
        oxides["total"] = oxides[oxides.elements].sum(axis=1)
        oxides._datatype = Datatype.OXIDE

        if units == Unit.MOL_FRACTIONS:
            if not normalise:
                return oxides
            return oxides.normalise()

        oxides_wt_pc = oxides.wt_pc(normalise=False)
        # oxides_wt_pc["total"] = oxides_wt_pc[oxides_wt_pc.elements].sum(axis=1)
        if not normalise:
            return oxides_wt_pc
        return oxides_wt_pc.normalise()

    @property
    def oxygen(self) -> pd.Series:
        """
        oxygen per 1 mole of cations
        """
        # Calculate oxide moles
        if self._datatype != Datatype.CATION:
            cations = self.cations()
            cations = cations[cations.elements]
        else:
            cations = self[self.elements].copy()

        oxygen_per_mole = oxide_compositions.oxygen_amount(
            cations.elements, type="cation"
        )
        cation_per_mole = oxide_compositions.cation_amount(
            cations.elements, type="cation"
        )

        oxygen_per_cation = oxygen_per_mole / cation_per_mole

        oxygen = cations.mul(oxygen_per_cation).sum(axis=1)

        return oxygen

    @_check_attribute("_units", ["wt. %", "ppm"])
    def convert_ppm_wtPercent(self) -> Self:
        """
        ppm converted to wt. % and vice versa
        """
        convert_dict = {
            Unit.WT_PERCENT: [1e4, Unit.PPM],
            Unit.PPM: [1e-4, Unit.WT_PERCENT],
        }

        converted = self.mul(convert_dict[self._units][0])
        converted._units = convert_dict[self._units][1]

        return converted

    @_check_attribute("_units", ["wt. %", "mol fraction"])
    def _convert_moles_wtPercent(self, normalise=True) -> Self:
        """
        moles converted to wt. % and vice versa
        """

        converted = self[self.elements].copy()
        if self._units == Unit.WT_PERCENT:
            converted = converted.div(converted.weights)
            units = Unit.MOL_FRACTIONS
        elif self._units == Unit.MOL_FRACTIONS:
            converted = converted.mul(converted.weights)
            units = Unit.WT_PERCENT

        if not normalise:
            converted["total"] = converted[converted.elements].sum(axis=1)
            converted._units = units
            return converted.recalculate()

        # Normalise
        total = converted[converted.elements].sum(axis=1)
        converted = converted.div(total, axis=0)
        converted["total"] = converted[converted.elements].sum(axis=1)
        # Set the right units

        if self._units == Unit.MOL_FRACTIONS:
            converted = converted.mul(100)

        converted._units = units

        return converted

    def mineral_formula(self, O: int = None) -> Self:
        """
        Calculate mineral formulas by normalising to oxygen per formula unit

        Parameters
        ----------
        O : int
            Amount of oxygen to normalise to.

        Returns
        -------
        mineral formulas : MagmaFrame
        """
        # Calculate cation fractions
        cations = self.cations()
        cations = cations[cations.elements]
        # Calculate oxygens per cation
        oxygen_numbers = e.oxygen_numbers(self.elements) / e.cation_numbers(
            self.elements
        )
        oxygen_numbers.index = cations.elements
        # Normalise to oxygen
        oxygen_total = cations.mul(oxygen_numbers).sum(axis=1)
        oxygen_factor = O / oxygen_total
        cations = cations.mul(oxygen_factor, axis=0)
        cations["O"] = O

        return cations

    def recalculate(self, inplace=False) -> Self:
        """
        Recalculate element masses and total weight.
        """
        df = self if inplace else self.copy()

        df._weights = element_weights.weights_as_series(self.columns)

        if df._total:
            totals = df.loc[:, df.elements].sum(axis=1)
            df.loc[:, "total"] = totals.values

        if not inplace:
            return df

    def normalise(self, to=None) -> Self:
        """
        Normalise compositions.

        Parameters
        ----------
        to :    float, int
            normalisation value

        Returns
        -------
        normalised data : MagmaFrame
        """
        if to is not None:
            norm = to
        elif self._units == Unit.WT_PERCENT:
            norm = 100
        else:
            norm = 1

        # self = self.recalculate()
        normalised = self[self.elements].copy()

        total = normalised.sum(axis=1)

        normalised = normalised.div(total, axis=0)
        normalised = normalised.mul(norm, axis=0)
        normalised.loc[:, "total"] = normalised.sum(axis=1)

        return normalised

    def random_sample(self, errors) -> Self:
        """
        Randomly resample compositions within errors.

        Sampling distribution is assumed normal with measured values as means and errors as standard deviations.

        Parameters
        ----------
        errors  : float, array-like
            standard deviation of the normal distributions. Use int for a fixed value for all elements or an array for specific values for all elements in :py:attr:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame.elements`

        Returns
        -------
        resampled data : MagmaFrame
            Randomly resampled compositions
        """

        random_sample = np.random.normal(self[self.elements], errors)
        random_sample[random_sample < 0] = 0

        df = self.copy()
        df[df.elements] = random_sample

        return df



import elementMass as e
import pandas as pd
from typing_extensions import Self

from MagmaPandas.enums import Datatype, Unit
from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame
from MagmaPandas.parse_io import check_components

_endmember_required_elements = ["MgO", "MnO", "FeO", "Fe2O3", "TiO2", "Al2O3"]


class Magnetite(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with magnetite specific methods.
    """

    def endmembers(self):
        """
        Calculations according to Lindsley, as implemented in QUILF
        """

        composition = check_components(
            composition=self, components=_endmember_required_elements
        )

        cations = composition.cations(norm_to=3)

        magnetite = (
            cations[["Mg", "Mn", "Fe"]].sum(axis=1)
            - 2 * cations["Ti"]
            - cations["Al"] / 2
        ) / 3
        mg_ulvospinel = cations["Mg"] / 2
        mn_ulvospinel = cations["Mn"] / 2
        ulvospinel = cations["Ti"] - mn_ulvospinel - mg_ulvospinel

        total = magnetite + ulvospinel + mg_ulvospinel
        ulvospinel = (ulvospinel + mg_ulvospinel) / total
        magnetite = 1 - ulvospinel

        return pd.DataFrame(
            {"magnetite": magnetite, "ulvospinel": ulvospinel}, index=self.index
        )

    def Fe_speciation(self) -> Self:
        """
        Calculations according to Lindsley, as implemented in QUILF
        """

        # TODO add code to catch when total iron is given as Fe2O3

        old_datatype = self._datatype
        elements = {Datatype.OXIDE: ("FeO", "Fe2O3"), Datatype.CATION: ("Fe", "Fe3")}[
            old_datatype
        ]

        if all(c in self.elements for c in elements):
            return self

        composition = check_components(
            composition=self, components=_endmember_required_elements
        )

        cations = composition.cations(normalise=False)

        totals = cations["total"]

        cations_norm = cations.div(totals, axis=0) * 3  # normalised to 3 cations total

        magnetite = (
            cations_norm[["Mg", "Mn", "Fe"]].sum(axis=1)
            - 2 * cations_norm["Ti"]
            - cations_norm["Al"] / 2
        ) / 3

        Fe3 = 2 * magnetite
        Fe2 = (
            magnetite
            + 2 * cations_norm["Ti"]
            + cations_norm["Al"] / 2
            - cations_norm["Mg"]
            - cations_norm["Mn"]
        )

        # cations_norm = cations_norm.drop(columns=["Fe"])
        cations_norm["Fe3"] = Fe3
        cations_norm["Fe"] = Fe2

        cations_Fe_calc = (cations_norm / 3).mul(totals, axis=0).recalculate()

        if old_datatype == Datatype.CATION:
            return cations_Fe_calc

        oxide_wt_pc = cations_Fe_calc.oxides(normalise=False).wt_pc(normalise=False)
        oxide_wt_pc["total"] = oxide_wt_pc[oxide_wt_pc.elements].sum(axis=1)

        # oxide_names = e.get_oxide_names(cations_Fe_calc.elements)
        # oxide_weights = e.compound_weights(oxide_names)
        # cation_amounts = e.cation_numbers(oxide_names)

        # oxide_wt_pc = cations_Fe_calc.rename(
        #     columns={
        #         cation: oxide
        #         for cation, oxide in zip(cations_Fe_calc.elements, oxide_names)
        #     }
        # ).recalculate()

        # oxide_wt_pc[oxide_wt_pc.elements] = (
        #     oxide_wt_pc[oxide_wt_pc.elements]
        #     .mul(oxide_weights, axis=1)
        #     .div(cation_amounts, axis=1)
        # )

        # oxide_wt_pc._units = Unit.WT_PERCENT
        # oxide_wt_pc._datatype = Datatype.OXIDE

        return oxide_wt_pc



# import elementMass as e
import pandas as pd
from typing_extensions import Self

from MagmaPandas.enums import Datatype, Unit
from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame
from MagmaPandas.parse_io import check_components

_endmember_required_elements = ["SiO2", "TiO2", "Al2O3", "FeO", "MnO", "MgO", "Cr2O3"]


class Ilmenite(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with ilmenite specific methods.
    """

    def endmembers(self):
        """
        Calculations according to Andersen et al. (1993), as implemented in QUILF
        """
        composition = check_components(
            composition=self, components=_endmember_required_elements
        )

        cations = composition.cations(norm_to=2)

        hematite_moles = (cations[["Fe", "Mg", "Mn"]].sum(axis=1) - cations["Ti"]) / 2

        Fe2 = cations["Ti"] - cations["Mg"] - cations["Mn"]

        ilmenite_moles = Fe2 + cations["Al"] / 2

        Fe3 = 2 * hematite_moles

        total_components = (
            hematite_moles + ilmenite_moles + cations[["Mn", "Mg"]].sum(axis=1)
        )

        hematite = hematite_moles / total_components
        ilmenite = ilmenite_moles / total_components
        geikielite = cations["Mg"] / total_components
        pyrophanite = cations["Mn"] / total_components

        return pd.DataFrame(
            {
                "hematite": hematite,
                "ilmenite": ilmenite,
                "geikielite": geikielite,
                "pyrophanite": pyrophanite,
            }
        )

    def Fe_speciation(self, normalise=False) -> Self:
        """
        Calculations according to Andersen et al. (1993), as implemented in QUILF
        """

        units = self._units
        datatype = self._datatype

        elements = {Datatype.OXIDE: ("FeO", "Fe2O3"), Datatype.CATION: ("Fe", "Fe3")}[
            datatype
        ]

        if all(c in self.elements for c in elements):
            return self

        composition = check_components(
            composition=self, components=_endmember_required_elements
        )

        cations = composition.cations(
            normalise=False
        )  # convert to cation mol fractions
        totals = cations["total"]
        cations = cations.div(totals, axis=0).mul(2, axis=0)

        hematite_moles = (cations[["Fe", "Mg", "Mn"]].sum(axis=1) - cations["Ti"]) / 2

        Fe2 = cations["Ti"] - cations["Mg"] - cations["Mn"]
        Fe3 = 2 * hematite_moles

        cations["Fe"] = Fe2
        cations["Fe3"] = Fe3
        cations = cations.recalculate()
        cations = cations.div(2, axis=0).mul(totals, axis=0)  # revert normalisation

        if units == Unit.MOL_FRACTIONS:
            if datatype == Datatype.CATION:
                return cations
            return cations.oxides(normalise=normalise)

        cations_wt_pc = cations.wt_pc(normalise=normalise)
        if datatype == Datatype.CATION:
            return cations_wt_pc

        return cations_wt_pc.oxides(normalise=normalise)



import numpy as np

from MagmaPandas.EOSs.parameters import EOSparams


def tait_VdP(phase, pkbar, T_K, Tref=298.15, **kwargs):
    """
    Pressure contribution to Gibb's free energy.
    Tait equation of state from Holland and Powell (2011)

    Parameters
    ----------
    phase       str
        Mineral phase
    pkbar       int, float
        Pressure in kilobar
    T_K           int, float
        Temperature in Kelvin
    tref        int, float
        Reference temperature in Kelvin

    Returns
    Pth         float
        Thermal pressure term
    a, b, c     float
        Equation of state parameters
    VdP         float
        Contribution of pressure to the Gibbs free energy
    """

    params = ["s", "v0", "n", "a0", "K0", "dKdP", "dKdP2"]
    s, v0, n, a0, K0, dKdP, dKdP2 = [getattr(EOSparams, phase)[i] for i in params]

    # Einstein temperature
    theta = 10636.0 / (s / n + 6.44)

    u0 = theta / Tref
    u = theta / T_K

    # Einstein function page 345
    xi0 = u0**2 * np.exp(u0) / (np.exp(u0) - 1) ** 2.0

    # Equation 3
    a = (1.0 + dKdP) / (1.0 + dKdP + K0 * dKdP2)
    b = dKdP / K0 - dKdP2 / (1.0 + dKdP)
    c = (1.0 + dKdP + K0 * dKdP2) / (dKdP**2.0 + dKdP - K0 * dKdP2)

    # thermal pressure term, equation 11
    Pth = a0 * K0 * theta / xi0 * (1 / (np.exp(u) - 1.0) - 1 / (np.exp(u0) - 1.0))

    # Intergral of volume, equation 13
    PV0 = pkbar * v0
    part1 = np.sign(1 - b * Pth) * abs(1 - b * Pth) ** (1 - c)
    part2 = np.sign(1 + b * (pkbar - Pth)) * abs(1 + b * (pkbar - Pth)) ** (1 - c)
    part3 = b * (c - 1) * pkbar

    VdP = PV0 * (1 - a + a * (part1 - part2) / part3)

    return VdP



"""
|CO2|-|H2O| solubility models from Allison et al. (2022)\ [32]_
"""

from typing import List, Tuple

import elementMass as e
import numpy as np
from scipy.constants import R
from scipy.optimize import root, root_scalar

from MagmaPandas.EOSs.CO2_H2O import hollowayBlank
from MagmaPandas.MagmaSeries.MagmaSeries_baseclass import MagmaSeries
from MagmaPandas.parse_io.validate import _check_argument, _check_setter
from MagmaPandas.volatile_solubility.solubility_baseclass import Solubility_model

"""
Equations from:

Allison, C. M., Roggensack, K., Clarke, A. B. (2022)
MafiCH: a general model for H2O–CO2 solubility in mafic magmas
Contributions to mineralogy and petrology 40

"""

fugacity_options = ["hollowayBlank"]
model_options = ["mixed", "h2o", "co2"]


class _meta_Allison_configuration(type):
    def __init__(cls, *args, **kwargs):
        cls._fugacity = "hollowayBlank"
        cls._model = "mixed"

    @property
    def fugacity(cls):
        return cls._fugacity

    @fugacity.setter
    @_check_setter(fugacity_options)
    def fugacity(cls, model: str):
        cls._fugacity = model

    @property
    def model(cls):
        return cls._model

    @model.setter
    @_check_setter(model_options)
    def model(cls, model: str):
        cls._model = model

    def __str__(cls):
        """ """

        variables = {"Fugacity model": "_fugacity", "Species model": "_model"}
        pad_left = 20
        pad_right = 20
        pad_total = pad_left + pad_right
        new_line = "\n"

        message = (
            f"{new_line}{' Allison (2022) volatile solubility ':#^{pad_total}}"
            f"{new_line}{'':#<{pad_total}}"
            f"{new_line}{'Settings':#<{pad_total}}"
        )

        parameter_settings = ""
        for param, model in variables.items():
            prameter_settings += (
                f"{new_line}{param:.<{pad_left}}{getattr(cls, model):.>{pad_right}}"
            )

        T_string = f"{1000+273.15:.0f}-{1400+273.14:.0f}\N{DEGREE SIGN}K"
        calibration_range = (
            f"{new_line}{'Calibration range':_<{pad_total}}"
            f"{new_line}{'Temperature':.<{pad_left}}{T_string:.>{pad_right}}"
            f"{new_line}{'Pressure':.<{pad_left}}{'< 7 kbar':.>{pad_right}}"
        )

        return message + parameter_settings + calibration_range

    def __str__(cls):
        variables = {"Fugacity model": "_fugacity", "Species model": "_model"}
        pad_left = 20
        pad_right = 20
        pad_total = pad_left + pad_right
        new_line = "\n"

        message = (
            f"{new_line}{' Allison (2022) volatile solubility ':#^{pad_total}}"
            f"{new_line}{'':#^{pad_total}}"
            f"{new_line}{'Settings':#<{pad_total}}"
        )
        parameter_settings = ""
        for param, model in variables.items():
            parameter_settings += (
                f"{param:.<{pad_left}}{getattr(cls, model):.>{pad_right}}"
            )
        temperature_string = f"{1000+273.15:.0f}-{1400+273.14:.0f}\N{DEGREE SIGN}K"
        calibration_range = (
            f"{new_line}{'Calibration_range':_<{pad_total}}"
            f"{new_line}{'Temperature':.<{pad_left}}{temperature_string:.>{pad_right}}"
            f"{new_line}{'Pressure':.<{pad_left}}{'< 7 kbar':.>{pad_right}}"
        )

        return message + parameter_settings + calibration_range


class Allison_configuration(metaclass=_meta_Allison_configuration):
    @classmethod
    def reset(cls):
        cls._fugacity = "hollowayBlank"
        cls._model = "mixed"

    # @classmethod
    # def print(cls):
    #     """ """

    #     variables = {"Fugacity model": "_fugacity", "Species model": "_model"}
    #     pad_left = 20
    #     pad_right = 20
    #     pad_total = pad_left + pad_right
    #     print(" Allison (2022) volatile solubility ".center(pad_total, "#"))
    #     print("".ljust(pad_total, "#"))
    #     print("Settings".ljust(pad_total, "_"))
    #     for param, model in variables.items():
    #         print(f"{param:.<{pad_left}}{getattr(cls, model):.>{pad_right}}")
    #     print("\nCalibration range".ljust(pad_total, "_"))
    #     T_string = f"{1000+273.15:.0f}-{1400+273.14:.0f}\N{DEGREE SIGN}K"
    #     print(f"{'Temperature':.<{pad_left}}{T_string:.>{pad_right}}")
    #     print(f"{'Pressure':.<{pad_left}}{'< 7 kbar':.>{pad_right}}")
    #     print("\n")


FeO_mass, Fe2O3_mass = e.compound_weights(["FeO", "Fe2O3"])
fugacity_model = hollowayBlank.fugacity


class h2o(Solubility_model):
    @staticmethod
    def calculate_saturation(
        oxide_wtPercents: MagmaSeries, T_K: float, **kwargs
    ) -> float:
        """
        Calculate melt |H2O| saturation pressure according to equation 8.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %. Needs to have an 'H2O' column.
        T_K : float
            temperature in Kelvin

        Returns
        -------
        P_saturation : float
            Saturation pressure in bar
        """
        x_fluid = kwargs.pop("x_fluid", 1.0)

        if oxide_wtPercents["H2O"] < 0:
            raise ValueError(f"H2O lower than 0: {oxide_wtPercents['H2O']}")
        if not (1 >= x_fluid >= 0):
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if oxide_wtPercents["H2O"] == 0.0:
            return 0.0

        composition = oxide_wtPercents.copy()

        H2O = composition["H2O"]
        fH2O = 104.98 * H2O**1.83
        fH2O_pure = fH2O / x_fluid

        P_saturation = root_scalar(
            _root_fugacity_pressure,
            args=(T_K, fH2O_pure, "H2O"),
            bracket=[1e-50, 1.5e4],
        ).root

        return P_saturation

    @staticmethod
    def calculate_solubility(P_bar: float, T_K: float, x_fluid=1.0, **kwargs) -> float:
        """
        Calculate melt |H2O| solubility according to equation 8.

        Parameters
        ----------
        P_bar : float
            Pressure in bar.
        T_K : float
            temperature in Kelvin
        x_fluid : float
            fraction of |H2O| in the fluid. Default value is 1.0

        Returns
        -------
        H2O : float
            melt |H2O| solubility in wt.%.
        """
        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        if any(i <= 0 for i in [P_bar, x_fluid]):
            return 0

        fH2O_pure = fugacity_model(T_K, P_bar, "H2O")
        fH2O = fH2O_pure * x_fluid

        return (fH2O / 104.98) ** (1 / 1.83)


class co2(Solubility_model):
    @staticmethod
    def calculate_saturation(
        oxide_wtPercents: MagmaSeries,
        T_K: float,
        **kwargs,
    ) -> float:
        """
        Calculate melt |CO2| saturation pressure according to equation 5.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %. Needs to have a 'CO2' column.
        T_K : float
            temperature in Kelvin

        Returns
        -------
        P_saturation : float
            Saturation pressure in bar
        """
        x_fluid = kwargs.pop("x_fluid", 0.0)

        if oxide_wtPercents["CO2"] < 0:
            raise ValueError(f"CO2 lower than 0: {oxide_wtPercents['CO2']}")
        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if oxide_wtPercents["CO2"] == 0.0:
            return 0.0

        composition = oxide_wtPercents.copy()
        CO2 = composition["CO2"]
        cations = co2._cation_fractions_Allison(composition)
        deltaV = co2._deltaV(cations)
        lnK0 = co2._lnK0(cations)
        FW = 36.594  # alkali basalt formula weight per 1 oxygen

        XCO3 = CO2 * (1 / 44.01) / ((100 / FW) - (CO2 / FW))
        Kf = XCO3 / (1 + XCO3)

        # Partial pressure of CO2
        P_CO2 = root_scalar(
            co2._root_partial_pressure,
            args=(T_K, deltaV, lnK0, Kf),
            bracket=[1e-50, 1.5e4],
        ).root

        if x_fluid <= 0:
            return P_CO2
        else:
            fCO2 = fugacity_model(T_K, P_CO2, "CO2")
            fCO2pure = fCO2 / (1 - x_fluid)

            # Saturation pressure
            P_saturation = root_scalar(
                _root_fugacity_pressure,
                args=(T_K, fCO2pure, "CO2"),
                bracket=[1e-50, 1.5e4],
            ).root

            return P_saturation

    @staticmethod
    def calculate_solubility(
        oxide_wtPercents: MagmaSeries,
        P_bar: float,
        T_K: float,
        x_fluid: float = 0.0,
        **kwargs,
    ) -> float:
        """
        Calculate melt |CO2| solubility according to equation 5.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %.
        P_bar : float
            Pressure in bar
        T_K : float
            temperature in Kelvin
        x_fluid : float
            fraction of |H2O| in the fluid. Default value is 0.0

        Returns
        -------
        solublities : float
            melt |CO2| solubility in wt. %.
        """

        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        if any(i <= 0 for i in [P_bar, (1 - x_fluid)]):
            return 0

        composition = oxide_wtPercents.copy()

        Ra = R * 10  # cm3.bar.K-1.mol-1
        FW = 36.594  # alkali basalt formula weight per 1 oxygen
        P0 = 1e3  # reference pressure in bars

        # CO2 fugacity of a pure fluid
        fCO2_pure = fugacity_model(T_K, P_bar, "CO2")
        # CO2 fugacity of the mixed fluid
        fCO2 = fCO2_pure * (1 - x_fluid)
        # Partial pressure of CO2
        if x_fluid > 0:
            P_CO2 = root_scalar(
                _root_fugacity_pressure, args=(T_K, fCO2, "CO2"), bracket=[1e-50, 1.5e4]
            ).root
        else:
            P_CO2 = P_bar

        cations = co2._cation_fractions_Allison(composition)
        deltaV = co2._deltaV(cations)
        lnK0 = co2._lnK0(cations)

        K = np.exp(lnK0) * np.exp(-deltaV * (P_CO2 - P0) / (Ra * T_K))
        Kf = K * fCO2
        XCO3 = Kf / (1 - Kf)
        CO2 = 44.01 * XCO3 / (44.01 * XCO3 + (1 - XCO3) * FW) * 100

        return CO2

    @staticmethod
    def _root_partial_pressure(P_bar, T_K, deltaV, lnK0, Kf):
        P0 = 1e3  # reference pressure
        Ra = R * 10  # Gas constant in cm3.bar.K-1.mol-1

        K_fugacity = Kf / fugacity_model(T_K, P_bar, "CO2")
        K_solubility = np.exp(lnK0) * np.exp(-deltaV * (P_bar - P0) / (Ra * T_K))
        return K_fugacity - K_solubility

    @staticmethod
    def _deltaV(cations):
        """ """
        NaK = cations["Na"] / (cations["Na"] + cations["K"])

        deltaV = (
            -3350.65
            + 2625.385 * cations["Ti"]
            + 3105.426 * cations["Al"]
            + 47.0037 * NaK
            + 3375.552 * np.sum([cations[i] for i in ["Si", "Na"]], axis=0)
            + 3795.115 * cations["K"]
            + 3628.018 * cations["Fe"]
            + 3323.32 * np.sum([cations[i] for i in ["Mg", "Ca"]], axis=0)
        )
        return deltaV

    @staticmethod
    def _lnK0(cations):
        """ """
        NaK = cations["Na"] / (cations["Na"] + cations["K"])

        lnK0 = (
            -128.365
            + 122.644 * np.sum([cations[i] for i in ["Fe", "Na", "Ca"]], axis=0)
            + 92.263 * np.sum([cations[i] for i in ["Ti", "Al"]], axis=0)
            + 114.098 * cations["Si"]
            + 111.549 * cations["Mg"]
            + 138.855 * cations["K"]
            + 2.239 * NaK
        )
        return lnK0

    @staticmethod
    def _cation_fractions_Allison(oxide_wtPercents):
        """ """
        elements = ["SiO2", "TiO2", "Al2O3", "FeO", "MgO", "CaO", "Na2O", "K2O"]

        composition = oxide_wtPercents.copy()

        # All Fe2O3 as Fe2+
        if "Fe2O3" in composition.index:
            mass_ratio = Fe2O3_mass / FeO_mass
            composition["FeO"] = composition["FeO"] + composition["Fe2O3"] / mass_ratio
        try:
            # For Series
            composition = composition.loc[elements]
        except KeyError:
            # For Dataframes
            composition = composition.loc[:, elements]
        composition.recalculate(inplace=True)

        cations = composition.cations()
        # Rounding to 3 decimals because Allison did the same
        # Results will be different if you don't
        cations = cations.round(3)

        return cations


class mixed(Solubility_model):
    @staticmethod
    @_check_argument("output", [None, "PXfl", "P", "Xfl"])
    def calculate_saturation(
        oxide_wtPercents: MagmaSeries,
        T_K: float | np.ndarray,
        output: str = "P",
        **kwargs,
    ) -> float | Tuple[float, float]:
        """
        Calculate volatile saturation pressure for systems with mixed |CO2|-|H2O| fluids.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composision in oxide wt. %. Needs to have 'H2O' and 'CO2' columns.
        T_K : float
            Temperature in kelvin
        output : str
            Output format. 'P' for pressure only, 'Xfl' for |H2O| fluid fraction only and 'PXfl' for both.

        Returns
        -------
        saturation : float, (float, float)
            Depending on the value of ``output``: saturation pressure in bar, |H2O| fluid fraction or (saturation pressure, fluid fraction)
        """
        composition = oxide_wtPercents.copy()

        P_H2O_saturation = h2o.calculate_saturation(composition, T_K=T_K, x_fluid=1.0)
        P_CO2_saturation = co2.calculate_saturation(composition, T_K=T_K, x_fluid=0.0)

        if oxide_wtPercents["H2O"] <= 0:
            return P_CO2_saturation
        if oxide_wtPercents["CO2"] <= 0:
            return P_H2O_saturation

        P_guess = 0

        for species in (P_H2O_saturation, P_CO2_saturation):
            if np.isfinite(species):
                P_guess += species

        saturation = root(
            mixed._saturation_rootFunction,
            x0=[P_guess, 0.1],
            args=(composition, T_K),
        ).x

        if saturation[1] <= 0.0:
            saturation[0] = P_CO2_saturation
        elif saturation[1] >= 1.0:
            saturation[0] = P_H2O_saturation
        saturation[1] = np.clip(saturation[1], 0.0, 1.0)

        return_dict = {"P": saturation[0], "Xfl": saturation[1], "PXfl": saturation}

        return return_dict[output]

    @staticmethod
    @_check_argument("output", [None, "both", "CO2", "H2O"])
    def calculate_solubility(
        oxide_wtPercents: MagmaSeries,
        P_bar: float,
        T_K: float,
        x_fluid: float,
        output: None | str = "both",
        **kwargs,
    ):
        """
        Calculate volatile solubilities for systems with mixed |CO2|-|H2O| fluids.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composision in oxide wt. %.
        P_bar : float
            pressure in bar
        T_K : float
            Temperature in kelvin
        x_fluid: float
            fraction of |H2O| in the fluid.
        output : str
            Output format. 'CO2' for |CO2| only, 'H2O' for |H2O| only and 'both' for both.

        Returns
        -------
        saturation : float, (float, float)
            Solubility in wt. %. Depending on the value of ``output``: |CO2|, |H2O| or (|CO2|, |H2O|).
        """

        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")

        composition = oxide_wtPercents.copy()

        H2O = h2o.calculate_solubility(P_bar, T_K, x_fluid)
        CO2 = co2.calculate_solubility(composition, P_bar, T_K, x_fluid)

        return_dict = {"both": np.array([H2O, CO2]), "H2O": H2O, "CO2": CO2}
        return return_dict[output]

    @staticmethod
    def _saturation_rootFunction(P_x_fluid: List, oxide_wtPercents: MagmaSeries, T_K):
        P_bar, x_fluid = P_x_fluid
        # Keep x_fluid and P_bar within bounds
        x_fluid = np.clip(x_fluid, 0.0, 1.0)
        P_bar = np.clip(P_bar, a_min=1e-15, a_max=None)

        composition = oxide_wtPercents.copy()

        H2O = composition["H2O"]
        CO2 = composition["CO2"]

        sample_concentrations = np.array([H2O, CO2])
        calculated_concentrations = np.array(
            mixed.calculate_solubility(
                oxide_wtPercents=composition,
                P_bar=P_bar,
                T_K=T_K,
                x_fluid=x_fluid,
            )
        )

        return abs(calculated_concentrations - sample_concentrations)


def _root_fugacity_pressure(P_bar, T_K, fugacity, species):
    return fugacity - fugacity_model(T_K=T_K, P_bar=P_bar, species=species)



import numpy as np

from MagmaPandas.EOSs.parameters import EOSparams
from MagmaPandas.EOSs.tait import tait_VdP


def landau(phase, pkbar, T_K, **kwargs):
    """
    Excess Gibbs free energy from Landau theory.

    Based on:
    Holland and Powell, 1998, p.312
    and
    Holland and Powell, 1990

    Rewritten according to Michael Anenberg:
    https://fo2.rses.anu.edu.au/fo2app/
    (see FMQ buffer details and references)

    Parameters
    ----------
    phase   str
        Mineral phase
    pkbar   int, float
        Pressure in kilobar
    t       int, float
        Temperature in Kelvin
    vmax    int, float
        Optional, maximum volume of disorder. Set to 0 to calculate the pressure independent contribution.

    Returns
    -------
    float
        Pressure dependent contribution to the excess Gibbs free enery from Landau theory
    """
    t = np.array([])
    t = np.append(t, T_K)

    vmax_default = getattr(EOSparams, phase)["vmax"]
    vmax = kwargs.get("vmax", vmax_default)

    smax, tc0 = [getattr(EOSparams, phase)[i] for i in ["smax", "Tc0"]]

    Q2_0 = np.sqrt(1 - 298.15 / tc0)

    # Landau critical temperature at pkbar
    tc = tc0 + pkbar * vmax / smax

    Q2 = np.zeros(
        shape=[
            len(t),
        ]
    )
    if any(t < tc):
        Q2 = np.where(t > tc, 0, np.sqrt((tc - t) / tc0))

    G_Landau = (
        smax * (tc0 * (Q2_0 + (Q2**3 - Q2_0**3) / 3) - tc * Q2 - t * (Q2_0 - Q2))
        + pkbar * vmax * Q2_0
    )

    # Convert to float if there is only one item
    if len(G_Landau) == 1:
        G_Landau = G_Landau.item()

    return G_Landau


def landau_P_dependent(phase, pkbar, T_K, formulation="anenberg"):
    """
    Pressure dependent excess Gibbs free energy from Landau theory.
    Calculated by subtracting the pressure independent contribution (with vmax = 0) from the total.

    Parameters
    ----------
    phase       :   str
        Mineral phase
    t           :   int, float
        Temperature in Kelvin
    pkbar       :   int, float
        Pressure in kilobars
    formulation :   str
        'holland' for the formulation by Holland and Powell (1998), otherwise Michael Anenberg's
        formulation will be used:
        https://fo2.rses.anu.edu.au/fo2app/
        (see FMQ buffer details and references)
    Returns
    -------
    float
        Pressure contribution to excess Gibbs free enery from Landau theory
    """
    if formulation == "holland":
        landau_total = landau_Holland(phase, pkbar, T_K)
        landau_1bar = landau_Holland(phase, 0, T_K, vmax=0)
    else:
        landau_total = landau(phase, pkbar, T_K)
        landau_1bar = landau(phase, 0, T_K, vmax=0)

    return landau_total - landau_1bar


def landau_Holland(phase, pkbar, T_K, **kwargs):
    """
    Excess Gibbs free energy from Landau theory

    Equations from Holland and Powell (1998), p. 312

    Parameters
    ----------
    phase   str
        Mineral phase
    pkbar   int, float
        Pressure in kilobars
    t       int, float
        Temperature in Kelvin

    Returns
    -------
    float
        Excess Gibbs free enery from Landau theory
    """
    t = np.array([])
    t = np.append(t, T_K)

    vmax_default = getattr(EOSparams, phase)["vmax"]
    vmax = kwargs.get("vmax", vmax_default)
    smax, tc0, a0, K0 = [
        getattr(EOSparams, phase)[i] for i in ["smax", "Tc0", "a0", "K0"]
    ]

    # Landau critical temperature
    tc = tc0 + vmax * pkbar / smax
    # Q: oder paramter in the landau model
    Q2_0 = np.sqrt(1 - 298.15 / tc0)

    # if t > tc:
    #     Q2 = 0
    # else:
    #     Q2 = np.sqrt((tc - t) / tc0)

    Q2 = np.where(t > tc, 0, np.sqrt((tc - t) / tc0))

    # Bulk modulus
    K = K0 * (1 - 1.5e-4 * (t - 298))

    # Excess enthalpy at 298K from Landau model disordering
    h = smax * tc0 * (Q2_0 - (Q2_0**3) / 3)
    # Excess entropy at 298K from Landau model disordering
    s = smax * Q2_0

    # Excess volume from Landau model disordering
    vt = vmax * Q2_0 * (1 + a0 * (t - 298)) - 20 * a0 * (np.sqrt(t) - np.sqrt(298))

    vtdP = vt * K / 3 * ((1 + 4 * pkbar / K) ** (3 / 4) - 1)

    delta_G_landau = smax * ((t - tc0) * Q2 + (tc * Q2**3) / 3)

    G_excess = h - t * s + vtdP + delta_G_landau

    # Convert to float if there is only one item
    if len(G_excess) == 1:
        G_excess = G_excess.item()

    return G_excess


def phase_transition(pkbar, T_K, phase_1, phase_2):
    """
    Gibbs free energy of transition from phase_1 to phase_2

    Parameters
    ----------
    pkbar               int, float
        Pressure in kilobar
    t                   int, float
        Temperature in Kelvin
    phase_1, phase_2    str
        Mineral phases

    Returns
    -------
    float
        Gibbs free energy of phase transition
    """

    results = []

    for phase in [phase_1, phase_2]:

        h = getattr(EOSparams, phase)["h"]
        s = getattr(EOSparams, phase)["s"] / 1e3

        Gibbs = (
            h
            + enthalpy(phase=phase, T_K=T_K)
            - T_K * (s + entropy(phase=phase, T_K=T_K))
        )
        VdP = tait_VdP(phase=phase, pkbar=pkbar, T_K=T_K)

        Gibbs = Gibbs + VdP

        if phase in ["quartz", "magnetite"]:

            Gibbs = Gibbs + landau(phase=phase, pkbar=pkbar, T_K=T_K)

        results.append(Gibbs)

    return results[0] - results[1]


def enthalpy(phase: str, T_K: int | float, Tref: int | float = 298.15):
    """
    Enthalpy as Cp.dT, integrated from T to T(reference), with heat capacity as:

    Cp = a + bT + cT**-2 + dT**(-1/2)

    Cp, a, b, c, & d from Holland and Powell (2011)

    Parameters
    ----------
    phase   str

    t       int, float
        temperature in Kelvin

    tref    int, float
        reference temperature in Kelvin

    Returns
    -------
    enthalpy   float
    """

    a, b, c, d = [
        getattr(EOSparams, phase)[i] for i in ["cp_a", "cp_b", "cp_c", "cp_d"]
    ]

    integral = lambda T: a * T + 0.5 * b * T**2.0 - c * T**-1.0 + 2 * d * T ** (1 / 2)
    enthalpy = integral(T_K) - integral(Tref)

    return enthalpy


def entropy(phase: str, T_K: int | float, Tref: int | float = 298.15):
    """
    Entropy as (Cp/T)dT, integrated from T to T(reference), with:

    Cp/T = a/T + b + cT**-3 + dT**(-3/2)

    Cp, a, b, c & d from Holland and Powell (2011)

    Parameters
    ----------
    phase   str

    t       int, float
        temperature in Kelvin

    tref    int, float
        reference temperature in Kelvin

    Returns
    -------
    entropy   float
    """

    a, b, c, d = [
        getattr(EOSparams, phase)[i] for i in ["cp_a", "cp_b", "cp_c", "cp_d"]
    ]

    integral = lambda T: a * np.log(T) + b * T - c / 2 * T**-2.0 - 2 * d * T ** (-1 / 2)
    entropy = integral(T_K) - integral(Tref)

    return entropy


def thermal_expansivity(V, V_0, alpha0, delta0, kappa):
    """
    Equation 5 from Komabayashi (2014)

    Parameters
    ----------
    V : float, array-like
        volume
    V_0 : float
        volume at reference conditions
    alpha0 : float
        thermal expansion at 1 bar
    delta0 : float
        Value of Anderson-Grüneisen parameter (delta_T) at 1 bar
    kappa : float
        dimensionless Anderson-Grüneisen parameter.


    Returns
    -------
    alpha
        thermal expansivity
    """

    return alpha0 * np.exp((-delta0 / kappa) * (1 - (V / V_0) ** kappa))



"""
Module with models for calculating melt Fe-Mg partition coefficients between olivine and melt as:

 (Fe\ :sup:`2+` / Mg)\ :sub:`ol` / (Fe\ :sup:`2+` / Mg)\ :sub:`melt`
"""

from MagmaPandas.Kd.Ol_melt.Kd_olmelt_FeMg_calculate import *
from MagmaPandas.Kd.Ol_melt.Kd_olmelt_FeMg_models import Kd_olmelt_FeMg_models_dict



import warnings as w
from multiprocessing import Pool

import numpy as np
import pandas as pd
from alive_progress import alive_bar
from typing_extensions import Self

from MagmaPandas import volatile_solubility as vs
from MagmaPandas.configuration import configuration
from MagmaPandas.enums import Datatype, Unit
from MagmaPandas.Fe_redox.Fe3Fe2_calculate import calculate_Fe3Fe2
from MagmaPandas.Kd.Ol_melt.Kd_olmelt_FeMg_models import Kd_olmelt_FeMg_models_dict
from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame
from MagmaPandas.parse_io.validate import _check_argument, _match_index
from MagmaPandas.rheology import calculate_density, calculate_viscosity
from MagmaPandas.thermometers import melt_thermometers


class Melt(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with melt specific methods.
    """

    # @property
    # def _constructor(self):
    #     """This is the key to letting Pandas know how to keep
    #     derivatives of `MagmaBase` the same type as yours.  It should
    #     be enough to return the name of the Class.  However, in
    #     some cases, `__finalize__` is not called and `new attributes` are
    #     not carried over.  We can fix that by constructing a callable
    #     that makes sure to call `__finalize__` every time."""

    #     def _c(*args, weights=None, **kwargs):
    #         if weights is None:
    #             weights = self._weights.copy(deep=True)
    #         return Melt(*args, weights=weights, **kwargs).__finalize__(self)

    #     return _c

    def temperature(self, P_bar: float | pd.Series = None, **kwargs) -> pd.Series:
        """
        Calculate melt liquidus temperatures.
        Model choice is set in the global :py:class:`~MagmaPandas.configuration.configuration` class.

        Parameters
        ----------
        P_bar   : float, pandas Series
            pressure in bar

        Returns
        -------
        temperatures : pd.Series
            Liquidus temperatures in Kelvin
        """
        thermometer = melt_thermometers[configuration.melt_thermometer]

        return thermometer(self.wt_pc(), P_bar=P_bar, **kwargs)

    def density(
        self,
        T_K: float | pd.Series,
        P_bar: float | pd.Series,
        fO2_logshift: None | int = None,
    ) -> pd.Series:
        """
        Calculate silicate melts densities with the Iacovino and Till (2019)\ [23]_ model

        Parameters
        ----------
        T_K : float, pandas Series
            temperatures in Kelvin
        P_bar : float, pandas Series
            pressures in bar
        fO2_logshift : None, int
            |fO2| buffer shift in log units. If set to None, the value set in the global configuration is used.

        Returns
        -------
        densities : pd.Series
            densities in kg/m\ :sup:`3`
        """

        _match_index(self, arg_names=("T_K", "P_bar"), kwargs=locals())

        if fO2_logshift is None:
            fO2_logshift = configuration.dfO2

        Fe3Fe2 = self.Fe3Fe2(T_K=T_K, P_bar=P_bar, fO2_logshift=fO2_logshift)
        # calculate melt composition with Fe2O3 and FeO
        melt = self.FeO_Fe2O3_calc(Fe3Fe2)

        return calculate_density(melt, T_K=T_K, P_bar=P_bar)

    def viscosity(self, T_K):
        """
        Calculate melts viscosity with the Giordano et al. (2008)\ [27]_ model

        Parameters
        ----------
        T_K : float, pandas Series
            temperatures in Kelvin

        Returns
        -------
        viscosities : pd.Series
            viscosity in log10(Pa.s)
        """

        return calculate_viscosity(melt_mol_fractions=self.moles(), T_K=T_K)

    def tetrahedral_cations(self) -> pd.Series:
        """
        Calculate tetrahedral cations based on Mysen (1983)\ [12]_

        Si, Ti, Al and P are assumed to be in tetrahedral coordination and Fe\ :sup:`3+` is not taken into account

        Returns
        -------
        tetrahedral cations : pd.Series
            summed tertrahedral cations per 1 mole cations

        """
        cations = self.cations()

        tetrahedral_cations = {"Si", "Ti", "Al", "P"}

        available_elements = tetrahedral_cations.intersection(cations.elements)

        return cations[list(available_elements)].sum(axis=1)

    def NBO(self):
        """
        Non-bridging oxygen in the melt
        Formulation according to Mysen (1983)\ [12]_

        Returns
        -------
        pd.Series
            NBO per 1 mole cations

        """

        oxygen = self.oxygen

        return (2 * oxygen) - (4 * self.tetrahedral_cations())

    def NBO_T(self):
        """
        NBO/T
        The ratio of non-bridging oxygen and tetrahedral cations according to Mysen (1983)\ [12]_

        Returns
        -------
        pd.Series
            NBO/T
        """

        return self.NBO() / self.tetrahedral_cations()

    def Fe3Fe2(
        self,
        T_K: float | pd.Series,
        P_bar: float | pd.Series,
        fO2_logshift: None | int = None,
        inplace=False,
        **kwargs,
    ) -> pd.Series:
        """
        Calculate melt |Fe3Fe2| ratios at the set |fO2| buffer.
        Model choice is set in the global :py:class:`~MagmaPandas.configuration.configuration` class.

        Parameters
        ----------

        T_K :   float, pd.Series-like
            temperatures in Kelvin
        Pbar    :   float, pd.Series-like
            Pressure in bars
        logshift    :   int, pd.Series-like
            |fO2| buffer shift in log units.
        inplace :   bool

        Returns
        -------
        melt Fe3+/Fe2+ ratios : pandas Series

        """
        # if T_K is None:
        #     T_K = self["T_K"]
        # if P_bar is None:
        #     P_bar = self["P_bar"]

        _match_index(self, arg_names=["T_K", "P_bar"], kwargs=locals())

        # for name in ["T_K", "P_bar"]:
        #     param = locals()[name]
        #     if isinstance(param, pd.Series):
        #         if not self.index.equals(param.index):
        #             raise RuntimeError(f"Melt and {name} indices don't match")

        if fO2_logshift is None:
            fO2_logshift = configuration.dfO2

        mol_fractions = self.moles()

        Fe3Fe2 = calculate_Fe3Fe2(
            mol_fractions=mol_fractions,
            T_K=T_K,
            P_bar=P_bar,
            dfO2=fO2_logshift,
            **kwargs,
        )

        if inplace:
            self["Fe3Fe2"] = Fe3Fe2
            if "T_K" not in self.columns:
                self["T_K"] = T_K
            if "Pbar" not in self.columns:
                self["Pbar"] = P_bar
        else:
            return Fe3Fe2

    @_check_argument("total_Fe", ["FeO", "Fe2O3"])
    def FeO_Fe2O3_calc(
        self,
        Fe3Fe2: float | pd.Series,
        total_Fe: str = "FeO",
        inplace: bool = False,
        wtpc=True,
    ) -> Self:
        """
        Calculate melt FeO and |Fe2O3| based on total Fe.

        Parameters
        ----------
        Fe3Fe2 : pandas Series
            melt |Fe3Fe2| ratios
        total_Fe    : str
            columname in Melt frame with total Fe
        inplace : bool

        Returns
        -------
        Melt    : Self
            melt compositions inclusding FeO and |Fe2O3|
        """

        Fe2Fe_total = 1 / (1 + Fe3Fe2)
        melt_mol_fractions = self.moles()

        if total_Fe == "FeO":
            Fe2 = melt_mol_fractions["FeO"] * Fe2Fe_total
            Fe3 = melt_mol_fractions["FeO"] * (1 - Fe2Fe_total) / 2
        if total_Fe == "Fe2O3":
            Fe2 = melt_mol_fractions["Fe2O3"] * Fe2Fe_total * 2
            Fe3 = melt_mol_fractions["Fe2O3"] * (1 - Fe2Fe_total)

        melt_mol_fractions["FeO"] = Fe2
        melt_mol_fractions["Fe2O3"] = Fe3
        melt_mol_fractions.recalculate(inplace=True)

        # Recalculate to wt. % (normalised)

        melt = melt_mol_fractions.wt_pc() if wtpc else melt_mol_fractions

        if inplace:
            self["FeO"] = melt["FeO"]
            self["Fe2O3"] = melt["Fe2O3"]
            self.recalculate(inplace=True)

        else:
            return melt

    def Kd_olivine_FeMg_eq(
        self,
        *args,
        **kwargs,
    ):
        """
        Calulate equilibrium Fe-Mg partitioning coefficients between olivine and melt as:

        (Fe\ :sup:`2+` / Mg)\ :sub:`ol` / (Fe\ :sup:`2+` / Mg)\ :sub:`melt`

        Model choice is set in the global :py:class:`~MagmaPandas.configuration.configuration` class.

        Parameters
        ----------
        forsterite : pandas Series
            initial olivine forsterite contents
        T_K : pandas Series
            temperatures in Kelvin
        kwargs
            Potential extra keyword arguments:

            #. 'T_K', temperature in Kelvin
            #. 'P_bar', pressure in bar
            #. 'Fe3Fe2', melt Fe\ :sup:`3+` /Fe\ :sup:`2+` ratios
            #. 'forsterite_initial', olivine forsterite content.

            Which extra keyword arguments are needed depends on the configured Kd model.

        Returns
        -------
        Kds :   pandas Series
            Fe-Mg partitioning coefficients
        """
        Kd_model_name = kwargs.get("Kd_model", configuration.Kd_model)
        Kd_model = Kd_olmelt_FeMg_models_dict[Kd_model_name]

        return Kd_model.calculate_Kd(
            melt_mol_fractions=self.moles(),
            *args,
            **kwargs,
        )

    def volatile_saturation_pressure(
        self, T_K: float | pd.Series, inplace: bool = False, **kwargs
    ):
        """
        Calculate melt volatile (|CO2| and/or |H2O|) saturation pressures.

        Model choice is set in the global :py:class:`~MagmaPandas.configuration.configuration` class.

        Parameters
        ----------
        T_K : float, pandas Series
            temperatures in Kelvin
        inplace : bool

        Returns
        -------
        P_bar   : pd.Series
            Saturation pressures in bar
        """

        if isinstance(T_K, pd.Series):
            T_K = T_K[self.index]
        elif isinstance(T_K, (float, int)):
            T_K = pd.Series(T_K, index=self.index)

        P_bar = pd.Series(index=self.index, dtype=float)
        Xfl = pd.Series(index=self.index, dtype=float)
        names = self.index.values

        model = configuration.volatile_solubility
        species = configuration.volatile_species

        samples = [
            (name, temperature, (model, species), kwargs)
            for name, temperature in zip(names, T_K)
        ]
        total = self.shape[0]

        # Run calculations in a pool across multiple cpu cores
        # Progress bar from alive_bar
        with (
            Pool() as pool,
            alive_bar(
                total,
                spinner=None,
                length=25,
                manual=True,
                theme="smooth",
                force_tty=True,
                title="Saturation pressure...",
                title_length=25,
            ) as bar,
        ):
            results = pool.imap_unordered(self._saturation_multicore, samples)
            pool.close()

            finished = 0
            bar(0.0 / total)

            for name, result in results:
                finished += 1
                bar(finished / total)
                try:
                    pressure, fl = result
                    P_bar[name] = pressure
                    Xfl[name] = fl
                except TypeError:
                    P_bar[name] = result

        if inplace:
            self["P_bar"] = P_bar
            return
        else:
            output = P_bar
            if sum(Xfl.isna()) != Xfl.shape[0]:
                output = [P_bar, Xfl]
            return output

    def _saturation_multicore(self, sample):
        """
        Refactor of calculate_saturation for multiprocess calling
        """

        name, temperature, model, kwargs = sample
        solubility_model = vs.get_solubility_model(*model)
        try:
            result = vs.calculate_saturation(
                self.loc[name],
                T_K=temperature,
                solubility_model=solubility_model,
                **kwargs,
            )
        except Exception as e:
            result = np.nan
            w.warn(f"Saturation pressure not found for sample {name}: {repr(e)}")

        return name, result

    # def _match_index(self, arg_names: list[str], kwargs):

    #     for name in arg_names:
    #         param = kwargs[name]
    #         if isinstance(param, pd.Series):
    #             if not self.index.equals(param.index):
    #                 raise RuntimeError(f"Melt and {name} indices don't match")



import elementMass as e
import matplotlib.pyplot as plt
import numpy as np

markers = ("<", "^", "s", "v", ">", "*", "p", ".", "P", "X", "8", "d")
marker_sizes = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.5, 1.2, 1.5, 1.2, 1.2, 1.0, 1.2])


# Color palettes
class colors:
    """
    Color palettes for plots
    """

    bright = plt.cycler(
        color=[
            "xkcd:bright lime green",
            "xkcd:dandelion",
            "xkcd:blood orange",
            "xkcd:white",
            "xkcd:ultramarine blue",
            "xkcd:light pink",
            "xkcd:green teal",
            "xkcd:neon yellow",
            "xkcd:cool blue",
            "xkcd:barbie pink",
            "xkcd:white",
            "xkcd:vivid purple",
        ],
        markeredgecolor=[
            "xkcd:forest",
            "xkcd:orange",
            "xkcd:white",
            "xkcd:blue green",
            "xkcd:pale cyan",
            "xkcd:easter purple",
            "xkcd:navy",
            "xkcd:off green",
            "xkcd:dark grey",
            "xkcd:deep violet",
            "xkcd:vibrant blue",
            "xkcd:pale lilac",
        ],
    )

    flatDesign = plt.cycler(
        color=["#e27a3d", "#344d5c", "#df5a49", "#43b29d", "#efc94d"]
    )

    firenze = plt.cycler(color=["#8E2800", "#468966", "#B64926", "#FFF0A5", "#FFB03B"])

    vitaminC = plt.cycler(color=["#FD7400", "#004358", "#FFE11A", "#1F8A70", "#BEDB39"])

    bella = plt.cycler(color=["#801637", "#047878", "#FFB733", "#F57336", "#C22121"])

    buddha = plt.cycler(color=["#192B33", "#FF8000", "#8FB359", "#FFD933", "#CCCC52"])

    elemental = plt.cycler(
        color=["#E64661", "#FFA644", "#998A2F", "#2C594F", "#002D40"]
    )

    carolina = plt.cycler(color=["#73839C", "#2E4569", "#AECCCF", "#D5957D", "#9C7873"])

    fourtyTwo = plt.cycler(
        color=["#2469A6", "#C4E1F2", "#F2E205", "#F2D22E", "#D9653B"]
    )

    terrazaverde = plt.cycler(
        color=["#DFE2F2", "#88ABF2", "#4384D9", "#56BFAC", "#D9B341"]
    )

    hollywood = plt.cycler(
        color=[
            "#8ECAE6",
            "#FD9E02",
            "#219EBC",
            "#FB8500",
            "#126782",
            "#FFB703",
            "#023047",
        ],
    )

    campfire = plt.cycler(color=["#588C7E", "#F2E394", "#F2AE72", "#D96459", "#8C4646"])

    pastel = plt.cycler(
        color=[
            "#FAD2E1",
            "#BCD4E6",
            "#C5DEDD",
            "#99C1DE",
            "#EDDCD2",
            "#DBE7E4",
            "#FFF1E6",
            "#F0EFEB",
            "#FDE2E4",
            "#D6E2E9",
        ],
    )

    autumn = plt.cycler(
        color=[
            "#797D62",
            "#FFCB69",
            "#9B9B7A",
            "#E8AC65",
            "#BAA587",
            "#D08C60",
            "#D9AE94",
            "#B58463",
            "#997B66",
            "#F1DCA7",
        ],
    )

    rainbow = plt.cycler(
        color=[
            "#54478C",
            "#83E377",
            "#2C699A",
            "#B9E769",
            "#048BA8",
            "#EFEA5A",
            "#0DB39E",
            "#F1C453",
            "#16DB93",
            "#F29E4C",
        ]
    )

    matteblue = plt.cycler(
        color=["#666A86", "#788AA3", "#92B6B1", "#B2C9AB", "#E8DDB5"],
    )


def layout(colors=colors.hollywood, fontsize=8, **kwargs):
    axTitleSize = int(fontsize)
    axLabelSize = int(fontsize)
    tickLabelSize = int(fontsize)
    markersize = kwargs.pop("markersize", 3)
    linewidth = kwargs.pop("linewidth", 1)
    facecolor = kwargs.pop("facecolor", "white")
    gridcolor = kwargs.pop("gridcolor", "whitesmoke")

    plt.rcParams["figure.constrained_layout.use"] = True

    plt.rc("figure", figsize=(8, 7), facecolor="white")

    # Text
    plt.rc("font", family="sans-serif", size=fontsize)

    # Legend
    plt.rc(
        "legend",
        fontsize=int(fontsize * 0.8),
        frameon=False,
        # facecolor="white",
    )

    # Axes
    plt.rc("xtick", direction="in", labelsize=tickLabelSize)
    plt.rc("ytick", direction="in", labelsize=tickLabelSize)
    plt.rc(
        "axes",
        grid=True,
        titlesize=axTitleSize,
        labelsize=axLabelSize,
        axisbelow=True,
        prop_cycle=colors
        + plt.cycler(
            marker=markers[: len(colors)],
            markersize=(marker_sizes * markersize)[: len(colors)],
        ),
        facecolor=facecolor,
        linewidth=0.8,
    )
    plt.rc("grid", color=gridcolor)

    # Lines
    plt.rc(
        "lines",
        markersize=markersize,
        linewidth=linewidth,
        markeredgewidth=0.5,
        markeredgecolor="k",
    )


def side_plots(
    ax,
    x_axis: bool = True,
    y_axis: bool = False,
    size=0.1,
    spacing=0,
    y_position="right",
    inside=False,
    **kwargs,
):
    axes = []

    if inside:
        spacing = -1

    if x_axis:
        ax_kde_x = ax.inset_axes([0, 1.0 + spacing, 1, size], sharex=ax)
        axes.append(ax_kde_x)

        ax_kde_x.spines["left"].set_visible(False)
        ax_kde_x.yaxis.set_visible(False)
        ax_kde_x.tick_params(axis="x", labelbottom=False, direction="in", **kwargs)
        # ax_kde_x.set_xlim(ax.get_xlim())
        ax_kde_x.spines["right"].set_visible(False)

    spacing = (spacing, -size)[inside]

    if y_axis:
        if y_position == "right":
            ax_position = ("right", "left")[inside]
            ax_kde_y = ax.inset_axes([1.0 + spacing, 0, size, 1], sharey=ax)
            ax_kde_y.spines[ax_position].set_visible(False)
            ax_kde_y.tick_params(
                axis="y",
                left=not inside,
                right=inside,
                labelright=False,
                direction="in",
                **kwargs,
            )
            if inside:
                ax_kde_y.invert_xaxis()

        if y_position == "left":
            ax_position = ("left", "right")[inside]
            ax_kde_y = ax.inset_axes([0 - spacing - size, 0, size, 1], sharey=ax)
            ax_kde_y.spines[ax_position].set_visible(False)
            ax_kde_y.tick_params(
                axis="y",
                left=inside,
                right=not inside,
                labelleft=False,
                direction="in",
                **kwargs,
            )
            if not inside:
                ax_kde_y.invert_xaxis()

        axes.append(ax_kde_y)

        ax_kde_y.tick_params(axis="y", labelleft=False, direction="in", **kwargs)
        ax_kde_y.xaxis.set_visible(False)
        ax_kde_y.spines["bottom"].set_visible(False)
        # ax_kde_y.tick_params(axis="y", labelleft=False, direction="in", **kwargs)
        # ax_kde_y.set_ylim(ax.get_ylim())

    for axis in axes:
        # axis.set_frame_on(False)
        axis.grid(False)
        axis.set_facecolor(color="white")

        for spine in ["top"]:
            axis.spines[spine].set_visible(False)

    return axes


def subscript_numbers(compound: str):
    string = ""
    elements = e.elements._find_elements(compound)
    for element in elements:
        name, quantity = e.elements._find_quantity(element)
        string = string + name
        if float(quantity) > 1:
            string = string + f"$_{quantity}$"

    return string



import numpy as np
import pandas as pd


def Kd_isopleths(
    kd_min: float = 0.1, kd_max: float = 0.4, stepsize: float = 0.1, **kwargs
):
    Kds = np.round(np.arange(kd_min, kd_max + 0.01, stepsize), 2)
    mg_no_melt_min, mg_no_melt_max, mg_no_melt_stepsize = kwargs.pop(
        "mg_no_melt", (0.2, 1, 0.01)
    )
    mg_no_melt = np.arange(mg_no_melt_min, mg_no_melt_max, mg_no_melt_stepsize)
    df = pd.DataFrame(index=mg_no_melt)

    Fe2Mg_melt = (1 - mg_no_melt) / mg_no_melt
    for kd in Kds:
        df[round(kd, 2)] = np.array(1 / (1 + (kd * Fe2Mg_melt))) * 100

    return df



import re
from typing import List, Optional

import elementMass as e
import numpy as np
import pandas as pd

from MagmaPandas.parse_io.validate import _check_argument


class Oxide_compositions:
    def __init__(self):
        self.names = np.array([], dtype=str)
        self._oxide_names = np.array([], dtype=str)
        self._cation_names = np.array([], dtype=str)
        self._cation_amount = np.array([], dtype=int)
        self._oxygen_amount = np.array([], dtype=int)

    def oxygen_amount(self, names, type="oxide"):
        if type == "oxide":
            self._process_names(names)
        idx = self._get_indeces(names, type=type)
        return self._oxygen_amount[idx]

    def cation_amount(self, names, type="oxide"):
        if type == "oxide":
            self._process_names(names)
        idx = self._get_indeces(names, type=type)
        return self._cation_amount[idx]

    def cation_names(self, oxides):
        self._process_names(oxides)
        idx = self._get_indeces(oxides)
        return self._cation_names[idx]

    def calculate_element_numbers(self, oxides):
        oxides_list = list(oxides)
        # new_oxides = oxides[~np.in1d(oxides, self.names)]
        # if len(new_oxides) == 0:
        #     return

        self.names = np.append(self.names, oxides_list)
        self._cation_names = np.append(self._cation_names, e.cation_names(oxides_list))
        self._cation_amount = np.append(
            self._cation_amount, e.cation_numbers(oxides_list).astype(int).values
        )
        self._oxygen_amount = np.append(
            self._oxygen_amount, e.oxygen_numbers(oxides_list).astype(int).values
        )

    def _process_names(self, oxides: List[str]) -> None:
        # cations = [c for c in oxides if "O" not in c]
        # oxides = list(set(oxides).difference(set(cations)))

        self._process_oxides(oxides=oxides)

        # difference = list(oxides_new.difference(set(self.names)))
        # if len(difference) > 0:
        #     self.calculate_element_numbers(difference)

        # idx = self._get_indeces(oxides)
        # [
        #     int(np.where(self.names == o)[0])
        #     for o in oxides[np.isin(oxides, self.names)]
        # ]

        # return idx

    def _process_oxides(self, oxides: List[str]) -> None:
        oxides_new = set(oxides)

        difference = list(oxides_new.difference(set(self.names)))
        if len(difference) > 0:
            self.calculate_element_numbers(difference)

    @_check_argument("type", [None, "oxide", "cation"])
    def _get_indeces(self, names: List[str], type: Optional[str] = None) -> List[int]:
        if type is None:
            type = "oxide"

        names_total = {"oxide": self.names, "cation": self._cation_names}[type]

        names_arr = np.array(names)

        return [
            int(np.where(names_total == name)[0])
            for name in names_arr[np.isin(names_arr, names_total)]
        ]


class Element_weights:
    def __init__(self):
        self.element_names = np.array([], dtype=str)
        self.all_weights = np.array([], dtype=float)
        self._not_an_element = np.array([], dtype=str)

    def weight(self, elements):
        idx = self._process_names(elements)
        return self.all_weights[idx].copy()

    def names(self, elements):
        idx = self._process_names(elements)
        return self.element_names[idx].copy()

    def calculate_weights(self, elements):
        new_elements = np.array(elements)[~np.in1d(elements, self.element_names)]
        if len(new_elements) == 0:
            return
        for element in new_elements:
            try:

                element_name = (
                    re.sub(r"\d+", "", element) if "O" not in element else element
                )  # strip numbers

                # Calculate element/oxide weight
                self.all_weights = np.append(
                    self.all_weights, e.calculate_weight(element_name)
                )
                self.element_names = np.append(self.element_names, element)
            except (ValueError, KeyError):
                self._not_an_element = np.append(self._not_an_element, element)

    def weights_as_series(self, elements):
        idx = self._process_names(elements)
        return pd.Series(self.all_weights[idx], index=self.element_names[idx])

    def _process_names(self, elements):
        elements = np.array(elements)

        difference = np.setdiff1d(
            elements, np.concatenate([self.element_names, self._not_an_element])
        )
        if len(difference) > 0:
            self.calculate_weights(difference)

        idx = [
            int(np.where(self.element_names == e)[0])
            for e in elements[np.isin(elements, self.element_names)]
        ]

        return idx


oxide_compositions = Oxide_compositions()
element_weights = Element_weights()



import pandas as pd
from typing_extensions import Self

from MagmaPandas.enums import Unit
from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame


class Clinopyroxene(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with clinopyroxene specific methods.
    """

    @property
    def formula(self) -> Self:
        """
        Mineral formulas normalised to 6 O p.f.u.
        """
        return self.mineral_formula(O=6)

    @property
    def endmembers(self) -> pd.DataFrame:
        """
        endmember components
        """

        cpx_formula = self.formula

        # TODO add code, check thermobar core.py calculate_clinopyroxene_components

        pass

    @property
    def mg_no(self) -> pd.Series:
        """
        Mg numbers
        """
        if self._units == Unit.WT_PERCENT:
            moles = self.moles()
        else:
            moles = self

        type = self._datatype.value

        Mg = {"oxide": "MgO", "cation": "Mg"}
        Fe = {"oxide": "FeO", "cation": "Fe"}
        self.recalculate(inplace=True)
        return pd.Series(
            moles[Mg[type]] / (moles[Fe[type]] + moles[Mg[type]]),
            name="Mg_no",
        )



from enum import Enum


class Unit(Enum):
    MOL_FRACTIONS = "mol fraction"
    WT_PERCENT = "wt. %"
    PPM = "ppm"
    UNKNOWN = None


class Datatype(Enum):
    CATION = "cation"
    OXIDE = "oxide"
    UNKNOWN = None



import itertools
from multiprocessing import Pool

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

import MagmaPandas.volatile_solubility as vs
from MagmaPandas.MagmaSeries import MagmaSeries


def CO2H2O_isobar_data(
    oxide_wtPercents: MagmaSeries,
    temperature: float,
    isobars=None,
    isopleths=None,
    interpolate=True,
    n_points=10,
):

    species = "mixed"

    if isobars is None:
        isobars = np.arange(1e3, 7e3, 1e3)

    if isopleths is None:
        isopleths = np.arange(0.0, 1.1, 0.1)

    isobar_data = pd.DataFrame(
        index=np.arange(0, 1 + 1 / n_points, 1 / n_points),
        columns=pd.MultiIndex.from_tuples(itertools.product(isobars, ["H2O", "CO2"])),
        dtype=float,
    )
    isobar_data.index.name = "xfl"
    isobar_data.columns.names = ("P_bar", "species")

    isopleth_data = pd.DataFrame(
        index=np.arange(0, max(isobars) + 1 / n_points, max(isobars) / n_points),
        columns=pd.MultiIndex.from_tuples(itertools.product(isopleths, ["H2O", "CO2"])),
        dtype=float,
    )
    isopleth_data.index.name = "P_bar"
    isopleth_data.columns.names = ("xfl", "species")

    samples_isobars = [
        pair
        for pair in itertools.product(
            [oxide_wtPercents],
            isobars,
            [temperature],
            isobar_data.index.values,
            [species],
        )
    ]
    samples_isopleths = [
        pair
        for pair in itertools.product(
            [oxide_wtPercents],
            isopleth_data.index.values,
            [temperature],
            isopleths,
            [species],
        )
    ]

    with Pool() as pool:

        results_isobars = pool.imap_unordered(_solubility_multicore, samples_isobars)

        for (xfl, P_bar), volatiles in results_isobars:
            isobar_data.loc[xfl, P_bar] = volatiles

        results_isopleths = pool.imap_unordered(
            _solubility_multicore, samples_isopleths
        )

        for (xfl, P_bar), volatiles in results_isopleths:
            isopleth_data.loc[P_bar, xfl] = volatiles

    if interpolate:
        return _isobar_interpolate(isobar_data, isopleth_data)

    return isobar_data, isopleth_data


def _solubility_multicore(sample):

    oxide_wtPercents, P_bar, temperature, xfl, species = sample

    volatiles = vs.calculate_solubility(
        oxide_wtPercents, P_bar=P_bar, T_K=temperature, x_fluid=xfl, species=species
    )

    return (xfl, P_bar), volatiles


def _isobar_interpolate(isobar_data, isopleth_data):

    isobars_smooth = pd.DataFrame(dtype=float, columns=["H2O", "CO2", "P_bar"])

    for p in isobar_data.columns.levels[0]:
        x_data = isobar_data[(p, "H2O")]
        y_data = isobar_data[(p, "CO2")]
        f = interp1d(x=x_data, y=y_data, kind="cubic")
        x_interp = np.linspace(min(x_data), max(x_data), len(x_data) * 10)
        data_interp = pd.DataFrame({"H2O": x_interp, "CO2": f(x_interp), "P_bar": p})
        isobars_smooth = pd.concat([isobars_smooth, data_interp])

    isopleths_smooth = pd.DataFrame(dtype=float, columns=["H2O", "CO2", "xfl"])

    trim = (0 < isopleth_data.columns.levels[0]) & (isopleth_data.columns.levels[0] < 1)

    for x in isopleth_data.columns.levels[0][trim]:
        x_data = isopleth_data[(x, "H2O")]
        y_data = isopleth_data[(x, "CO2")]
        f = interp1d(x=x_data, y=y_data, kind="quadratic")
        x_interp = np.linspace(min(x_data), max(x_data), len(x_data) * 10)
        data_interp = pd.DataFrame({"H2O": x_interp, "CO2": f(x_interp), "xfl": x})
        isopleths_smooth = pd.concat([isopleths_smooth, data_interp])

    if 0.0 in isopleth_data.columns.levels[0]:
        CO2max = isobars_smooth.loc[isobars_smooth["H2O"] == 0.0, "CO2"].max()
        xfl0 = pd.DataFrame({"H2O": (0, 0), "CO2": (0, CO2max), "xfl": (0.0, 0.0)})
        isopleths_smooth = pd.concat([isopleths_smooth, xfl0], axis=0)

    if 1.0 in isopleth_data.columns.levels[0]:
        H2Omax = isobars_smooth.loc[isobars_smooth["CO2"] == 0.0, "H2O"].max()
        xfl1 = pd.DataFrame({"H2O": (0, H2Omax), "CO2": (0.0, 0.0), "xfl": (1.0, 1.0)})
        isopleths_smooth = pd.concat([isopleths_smooth, xfl1], axis=0)

    return isobars_smooth, isopleths_smooth



from typing import List

import elementMass as e
import pandas as pd

from MagmaPandas import MagmaFrames
from MagmaPandas.MagmaFrames import (
    Clinopyroxene,
    MagmaFrame,
    Melt,
    Olivine,
    Plagioclase,
)
from MagmaPandas.parse_io.validate import _check_argument


@_check_argument("phase", [None, "Melt", "Olivine", "Clinopyroxene", "Plagioclase"])
def read_file(
    filepath: str,
    *args,
    total_col: str,
    index_col: List[str] = None,
    keep_columns: List[str] = None,
    phase: str = None,
    units: str = None,
    datatype: str = None,
    **kwargs,
) -> MagmaFrame:
    """
    Read compositions from a .csv file into a pd.dataframe, clean the data and pass it to a magmaframe.

    Parameters
    ----------
    filepath    :   str
        path to a .csv file with sample compositions
    phase   :       str


    """
    if phase is None:
        phase = "MagmaFrame"
    if keep_columns is None:
        keep_columns = []

    df = pd.read_csv(filepath, *args, index_col=index_col, **kwargs)

    delete_columns = set()
    elements = []

    # Check which column names have valid names for elements or oxides
    for col in df.columns:
        try:
            _ = e.calculate_weight(col)
            elements.append(col)
        except:
            delete_columns.add(col)

    if total_col is not None:
        df.rename(columns={total_col: "total"})
        df["total"] = df[elements].sum(axis=1)

    if "total" in df.columns:
        keep_columns.append("total")

    # Drop all columns without chemical data, unless explicitely specified otherwise in 'keep_columns'
    df = df.drop(delete_columns.difference(set(keep_columns)), axis=1)
    df[elements] = df[elements].astype("float32")

    create_class_instance = getattr(MagmaFrames, phase)

    return create_class_instance(df, units=units, datatype=datatype)


def read_clinopyroxene(
    filepath: str,
    *args,
    index_col: List[str] = None,
    total_col: str = None,
    keep_columns: List[str] = None,
    **kwargs,
) -> Clinopyroxene:
    """
    Read cpx compositions in wt. % oxide from a .csv file

    """

    return read_file(
        filepath=filepath,
        *args,
        phase="Clinopyroxene",
        index_col=index_col,
        total_col=total_col,
        keep_columns=keep_columns,
        units="wt. %",
        datatype="oxide",
        **kwargs,
    )


def read_melt(
    filepath: str,
    *args,
    index_col: List[str] = None,
    total_col: str = None,
    keep_columns: List[str] = None,
    units="wt. %",
    datatype="oxide",
    **kwargs,
) -> Melt:
    """
    Read melt compositions in wt. % oxide from a .csv file

    """

    return read_file(
        filepath=filepath,
        *args,
        phase="Melt",
        index_col=index_col,
        total_col=total_col,
        keep_columns=keep_columns,
        units=units,
        datatype=datatype,
        **kwargs,
    )


def read_olivine(
    filepath: str,
    *args,
    index_col: List[str] = None,
    total_col: str = None,
    keep_columns: List[str] = None,
    **kwargs,
) -> Olivine:
    """
    Read olivine compositions in wt. % oxide from a .csv file

    """

    return read_file(
        filepath=filepath,
        *args,
        phase="Olivine",
        index_col=index_col,
        total_col=total_col,
        keep_columns=keep_columns,
        units="wt. %",
        datatype="oxide",
        **kwargs,
    )


def read_plagioclase(
    filepath: str,
    *args,
    index_col: List[str] = None,
    total_col: str = None,
    keep_columns: List[str] = None,
    **kwargs,
) -> Plagioclase:
    """
    Read plagioclase compositions in wt. % oxide from a .csv file

    """

    return read_file(
        filepath=filepath,
        *args,
        phase="Plagioclase",
        index_col=index_col,
        total_col=total_col,
        keep_columns=keep_columns,
        units="wt. %",
        datatype="oxide",
        **kwargs,
    )



import inspect
import warnings as w
from functools import wraps
from typing import List

import numpy as np
import pandas as pd


def _check_argument(var_name: str, allowed_values: List[str]):
    def decorator(func):
        """
        Check if var_name has a valid value
        """
        # Get values for default arguments
        sig = inspect.signature(func)
        defaults = {
            n: val
            for n, val in zip(
                sig.parameters.keys(), [val.default for val in sig.parameters.values()]
            )
        }

        @wraps(func)  # comment
        def wrapper(*args, **kwargs):

            for key in kwargs.keys():
                # Remove user kwargs from the default dict
                defaults.pop(key, None)

            var = {**kwargs, **defaults}.get(var_name, None)
            if var not in allowed_values:
                raise ValueError(
                    f"{var_name}: {var}, not recognised, please choose from: {allowed_values}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _check_setter(allowed_values):
    def decorator(func):
        """
        check if set value is valid
        """

        @wraps(func)  # comment
        def wrapper(*args):
            var = args[1]
            if isinstance(var, (float, int)):
                min, max = allowed_values
                if not (min < var < max):
                    raise ValueError(
                        f"value: {var}, outside allowed range: {*allowed_values,}"
                    )
            elif isinstance(var, str):
                if var not in allowed_values:
                    raise ValueError(
                        f"'{var}' is not recognised, please choose from: {*allowed_values,}"
                    )
            return func(*args)

        return wrapper

    return decorator


def _check_attribute(attr_name: str, allowed_values: List[str]):
    def decorator(func):
        """
        Check if attr_name has a valid value
        """

        @wraps(func)  # comment
        def wrapper(self, *args, **kwargs):
            attr = getattr(self, attr_name)
            if attr.value not in allowed_values:
                raise ValueError(
                    f"Calculation is not valid with: {attr.value}, please use: {*allowed_values,}"
                )
            return func(self, *args, **kwargs)

        # wrapper.__doc__ = func.__doc__
        return wrapper

    return decorator


def _check_value(var_name: str, allowed_range: List[float], error=True):
    def decorator(func):
        """
        Check if var_name has a valid value
        """

        @wraps(func)  # comment
        def wrapper(*args, **kwargs):
            var = kwargs.get(var_name, None)
            xmin, xmax = allowed_range
            try:
                # ints and floats
                within_range = xmin < var < xmax
            except ValueError:
                # array-likes
                within_range = np.any((var > xmin) & (var < xmax))
            if not within_range:
                if error:
                    raise ValueError(
                        f"(some) values of {var_name} are outside allowed range: {xmin:.3f} - {xmax:.3f}"
                    )
                else:
                    w.warn(
                        f"(some) values of {var_name} are outside allowed range: {xmin:.3f} - {xmax:.3f})"
                    )
            # if not (min < var < max):
            #     raise ValueError(
            #         f"{var_name}: {var}, outside allowed range: {min} - {max}"
            #     )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _match_index(x, arg_names: list[str], kwargs):

    for name in arg_names:
        param = kwargs[name]
        if not isinstance(param, (pd.Series, pd.DataFrame)):
            continue
        if not x.index.equals(param.index):
            raise RuntimeError(f"index doesn't match with {name}")



import warnings as w

import numpy as np
import pandas as pd
import scipy.optimize as opt
from scipy.constants import R

from MagmaPandas.EOSs.tait import tait_VdP
from MagmaPandas.EOSs.tools import landau_P_dependent, phase_transition
from MagmaPandas.parse_io import repeat_vars


def VdP_QFM(T_K, P_bar):
    """
    Solve Tait equations of state for VdP of quartz, magnetite and fayalite
    at temperature, T_K and P_bar, ignoring phase transitions.
    """

    p_kbar = P_bar / 1e3

    # VdP SiO2
    VdP_qtz = tait_VdP(phase="quartz", pkbar=p_kbar, T_K=T_K)
    Gibbs_landau = landau_P_dependent(phase="quartz", pkbar=p_kbar, T_K=T_K)
    VdP_qtz = VdP_qtz + Gibbs_landau

    VdP_fay = tait_VdP(phase="fayalite", pkbar=p_kbar, T_K=T_K)

    VdP_mt = tait_VdP(phase="magnetite", pkbar=p_kbar, T_K=T_K)

    return VdP_qtz, VdP_mt, VdP_fay


def VdP_QFM_phaseTransitions(T_K, P_bar):
    """
    Solve Tait equations of state for VdP for quartz, magnetite and fayalite
    at temperature, T_K and P_bar, taking into account phase transitions
    """

    p_kbar = P_bar / 1e3

    if isinstance(p_kbar, pd.Series):
        p_kbar = p_kbar.squeeze()
    if isinstance(T_K, pd.Series):
        T_K = T_K.squeeze()

    # Calculate pressure of transition for SiO2 polymorphs
    # Quartz --> coesite
    qtz_coe = lambda P: phase_transition(
        P, T_K=T_K, phase_1="quartz", phase_2="coesite"
    )
    P_qtz_coe = opt.fsolve(qtz_coe, 8)
    # Coesite --> stishovite
    coe_stish = lambda P: phase_transition(
        P, T_K=T_K, phase_1="coesite", phase_2="stishovite"
    )
    P_coe_stish = opt.fsolve(coe_stish, 8)

    # Calculate pressure of transition for polymorphs
    # Fayalite --> ringwoodite
    fay_ring = lambda P: phase_transition(
        P, T_K=T_K, phase_1="fayalite", phase_2="ringwoodite"
    )
    P_fay_ring = opt.fsolve(fay_ring, 8)

    # Integrate VdP for all polymorphs
    # SiO2 polymorphs
    # Quartz
    VdP_SiO2 = tait_VdP(phase="quartz", pkbar=min(p_kbar, P_qtz_coe), T_K=T_K)
    SiO2_landau = landau_P_dependent(
        phase="quartz", pkbar=min(p_kbar, P_qtz_coe), T_K=T_K
    )
    VdP_SiO2 = VdP_SiO2 + SiO2_landau
    if p_kbar > P_qtz_coe:
        # Coesite
        VdP_coe = tait_VdP(
            phase="coesite", pkbar=min(p_kbar, P_coe_stish), T_K=T_K
        ) - tait_VdP(phase="coesite", pkbar=P_qtz_coe, T_K=T_K)
        VdP_SiO2 = VdP_SiO2 + VdP_coe
        if p_kbar > P_coe_stish:
            # Stishovite
            VdP_stish = tait_VdP(phase="stishovite", pkbar=p_kbar, T_K=T_K) - tait_VdP(
                phase="stishovite", pkbar=P_coe_stish, T_K=T_K
            )
            VdP_SiO2 = VdP_SiO2 + VdP_stish

    # Fe2SiO4 polymorphs
    # Fayalite
    VdP_Fe2SiO4 = tait_VdP(phase="fayalite", pkbar=min(p_kbar, P_fay_ring), T_K=T_K)
    if p_kbar > P_fay_ring:
        # Ringwoodite
        VdP_ring = tait_VdP(phase="ringwoodite", pkbar=p_kbar, T_K=T_K) - tait_VdP(
            phase="ringwoodite", pkbar=P_fay_ring, T_K=T_K
        )
        VdP_Fe2SiO4 = VdP_Fe2SiO4 + VdP_ring

    # Magnetite
    VdP_mt = tait_VdP(phase="magnetite", pkbar=p_kbar, T_K=T_K)

    return VdP_SiO2, VdP_mt, VdP_Fe2SiO4


def muO2_QFM_P(T_K, P_bar):
    """
    calculate chemical potential of oxygen at QFM and pressure P with equations of state
    """

    P_bar_is_int = True if isinstance(P_bar, (int, float)) else False
    T_K_is_int = True if isinstance(T_K, (int, float)) else False

    # If P and T are both single values
    if P_bar_is_int and T_K_is_int:
        VdP_quartz, VdP_magnetite, VdP_fayalite = VdP_QFM_phaseTransitions(T_K, P_bar)
        muO2 = 1e3 * (3 * VdP_quartz + 2 * VdP_magnetite - 3 * VdP_fayalite)
        return muO2

    # repeat the short variable
    T_K, P_bar = repeat_vars(T_K, P_bar)

    muO2 = np.array([])

    for temperature, pressure in zip(T_K, P_bar):
        VdP_quartz, VdP_magnetite, VdP_fayalite = VdP_QFM_phaseTransitions(
            temperature, pressure
        )
        # kiloJoule to Joule
        muO2 = np.append(
            muO2, 1e3 * (3 * VdP_quartz + 2 * VdP_magnetite - 3 * VdP_fayalite)
        )

    return muO2


def muO2_QFM_1bar(T_K, warning=False):
    """
    calculate chemical potential of oxygen at QFM a 1 bar. Equation from O'Neill 1987

    Parameters
    ----------
    T_K     list-like, float
        Temperature in Kelvin

    Returns
    -------
    Chemical potential
    """

    T = np.array([])
    T = np.append(T, T_K)

    if warning:
        if (np.array(T_K) < 900).any():
            w.warn("O'Neill fO2: temperatures below 900K present")
        if (np.array(T_K) > 1420).any():
            w.warn("O'Neill fO2: temperatures above 1420K present")

    muO2 = -587474 + 1584.427 * T - 203.3164 * T * np.log(T) + 0.092710 * T**2

    # Convert to float if there is only one item
    if len(muO2) == 1:
        muO2 = muO2.item()

    return muO2


def fO2_QFM_1bar(T_K, logshift=0):
    """
    calculate fO2 at QFM + logshift a 1 bar. Equation from O'Neill 1987

    Parameters
    ----------
    T_K     list-like, float
        Temperature in Kelvin
    logshift   int, float
        Log units by which QFM is shifted

    Returns
    -------
    fO2
    """
    mu_O2 = muO2_QFM_1bar(T_K)

    offset = 10**logshift

    return np.exp(mu_O2 / (R * T_K)) * offset


def fO2_QFM(
    logshift: int | float, T_K: float | np.ndarray, P_bar: float | np.ndarray
) -> float | np.ndarray:
    """
    Calculate |fO2| at the QFM buffer.

    1 bar components is calculated according to O'Neill (1987)\ [13]_ and pressure contributions according to Holland and Powell (2011)\ [14]_, with Landau theory from Holland and Powell (1990, 1998)\ [15]_:sup:`,`\ [16]_ and thermal Tait equations of state parameters from Holland and Powell (2011)\ [14]_, updated by Jennings and Holland (2015)\ [17]_.

    Parameters
    ----------
    logshift    : int, float
        |fO2| buffer shift in log units of QFM.
    T_K : float, array-like
        temperatures in Kelvin
    P_bar : float, array-like
        pressure in bar

    Returns
    -------
    fO2 : float, array-like
        |fO2| in bar
    """
    try:
        logshift = float(logshift)
    except TypeError:
        pass

    offset = 10**logshift

    # Chemical potential of oxygen
    # 1 bar contribution from O'Neill
    muO2_1bar_Oneill = muO2_QFM_1bar(T_K)
    # Pressue contribution from equations of state
    muO2_pressure_eos = muO2_QFM_P(T_K, P_bar)
    # Remove any 1 bar contribution from the equation of state formulation,
    # since the O'Neill emperical formulation is used for this
    VdP_quartz, VdP_magnetite, VdP_fayalite = VdP_QFM(T_K, 1)
    muO2_1bar_eos = 1e3 * (3 * VdP_quartz + 2 * VdP_magnetite - 3 * VdP_fayalite)
    muO2_pressure = muO2_pressure_eos - muO2_1bar_eos
    # Total chemical potential
    muO2 = muO2_1bar_Oneill + muO2_pressure

    fO2 = np.exp(muO2 / (R * T_K)) * offset

    try:
        fO2 = np.float32(fO2.item())
    except ValueError:
        fO2 = fO2.astype(np.float32)

    # if isinstance(fO2, pd.Series):
    #     return fO2.squeeze()

    return fO2



import warnings as w
from fractions import Fraction
from typing import Dict, List

import elementMass as e
import numpy as np
import pandas as pd

from MagmaPandas.MagmaFrames.protocols import Magma


def _check_calibration_range_Series(melt: pd.Series, calibration_range: Dict) -> None:

    checks = np.array([], dtype=bool)
    for element, *range in calibration_range:
        row = [element] if isinstance(element, str) else element
        np.append(checks, not (range[0] < melt[row].sum() < range[1]))

    if sum(checks) < 1:
        return

    w.warn(f"sample has composition outside the thermometer calibration range")

    # SiO2_range = not (31 < melt["SiO2"] < 73.64)
    # alkali_range = melt[["Na2O", "K2O"]].sum() > 14.3
    # H2O_range = melt["H2O"] > 18.6

    # outside_range = SiO2_range | alkali_range | H2O_range

    # if not outside_range:
    #     return
    # w.warn(f"sample has composition outside the thermometer calibration range")


def _check_calibration_range_dataframe(melt: Magma, calibration_range: Dict) -> None:

    checks = pd.DataFrame(dtype=bool)

    for element, *range in calibration_range:
        col = [element] if isinstance(element, str) else element
        checks[str(element)] = ~melt[col].sum(axis=1).between(range[0], range[1])

    outside_range = checks.any(axis=1)

    if outside_range.sum() < 1:
        return

    w.warn(
        f"samples {*melt.index[outside_range],} have compositions outside the thermometer calibration range"
    )

    # SiO2_range = ~melt["SiO2"].between(31, 73.64)
    # alkali_range = melt[["Na2O", "K2O"]].sum(axis=1) > 14.3
    # H2O_range = melt["H2O"] > 18.6

    # outside_range = SiO2_range | alkali_range | H2O_range

    # if outside_range.sum() < 1:
    #     return

    # w.warn(
    #     f"samples {*melt.index[outside_range],} have compositions outside the thermometer calibration range"
    # )


def _check_calibration_range(melt: pd.Series | Magma, calibration_range) -> None:

    if isinstance(melt, pd.Series):
        _check_calibration_range_Series(melt, calibration_range)
    elif isinstance(melt, pd.DataFrame):
        _check_calibration_range_dataframe(melt, calibration_range)


def _get_elements(composition):

    import MagmaPandas as mp

    if isinstance(composition, mp.MagmaFrame):
        return composition.columns

    if isinstance(composition, mp.MagmaSeries):
        return composition.index


def _anhydrous_composition(composition):

    composition_H2O = composition.copy()

    try:
        composition_H2O = composition_H2O.drop("H2O")
    except KeyError:
        composition_H2O = composition_H2O.drop(columns=["H2O"])

    return composition_H2O.recalculate().normalise()


def _remove_elements(composition, drop: List[str]):

    composition_new = composition.copy()

    for e in drop:

        if e in composition_new.index:
            composition_new = composition_new.drop(e)
        try:
            if e in composition_new.columns:
                composition_new = composition_new.drop(columns=[e])
        except AttributeError:
            pass

        composition_new.recalculate(inplace=True)

    return composition_new


def moles_per_oxygen(moles):

    axis = [0, 1][isinstance(moles, pd.DataFrame)]

    O_number = e.oxygen_numbers(moles.elements)

    no_oxygen = O_number.index[O_number == 0]
    composition_new = _remove_elements(composition=moles, drop=no_oxygen)

    O_number = O_number.drop(no_oxygen)
    cations = e.cation_names(composition_new.elements)
    cation_number = e.cation_numbers(composition_new.elements)

    cations_normalised = cation_number.div(O_number)

    names_new = {
        ox: f"{cat}{Fraction(num).limit_denominator()}O"
        for cat, num, ox in zip(cations, cations_normalised, composition_new.elements)
    }
    names_new["total"] = "O_total"

    composition_new = (
        composition_new[composition_new.elements]
        .mul(O_number, axis=axis)
        .normalise(to=1)
    )

    return composition_new.rename(names_new, axis=axis)


def _check_temperature(T_K):
    """
    Find negative or NaN temperatures
    """
    try:
        if (neg := any(T_K < 0)) or (nan := any(np.isnan(T_K))):
            str_arr = np.array(["negative", "NaN"])[np.array([neg, nan])]
            w.warn(f"{', '.join(str_arr)} temperatures found!")
    except TypeError:
        if (neg := (T_K < 0)) or (nan := (T_K != T_K)):
            str_arr = np.array(["negative", "NaN"])[np.array([neg, nan])]



import inspect
import sys
from functools import partial

import numpy as np
import pandas as pd
from scipy.constants import R  # J*K-1*mol-1

from MagmaPandas import configuration
from MagmaPandas.Fe_redox import Fe3Fe2_models
from MagmaPandas.fO2 import calculate_fO2
from MagmaPandas.Kd.Kd_baseclass import Kd_model
from MagmaPandas.Kd.Ol_melt.Kd_olmelt_FeMg_iterate import (
    iterate_Kd_scalar,
    iterate_Kd_vectorized,
)
from MagmaPandas.parse_io import check_components
from MagmaPandas.thermometers.data_parsing import _remove_elements, moles_per_oxygen


def _is_Kd_model(cls):
    """
    check if class inherits from Kd_model
    """
    try:
        return isinstance(cls(), Kd_model)
    except TypeError:
        return False


class toplis(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 10 from Toplis (2005)\ [19]_.
    """

    error = 0.02

    @classmethod
    def _Phi(cls, molar_SiO2, molar_Na2O, molar_K2O):
        """
        Equation 12 from Toplis (2005) calculates a Phi parameter to correct SiO2 for alkali-bearing liquids.

        Parameters
        ----------
        molar_SiO2 : int or list-like

        molar_Na2O : int or list-like

        molar_K2O : int or list-like


        Returns
        -------
            int or list-like
        """

        try:
            if sum(high_SiO2 := np.array(molar_SiO2) > 60) > 1:
                # raise RuntimeError("SiO2 >60 mol% present")
                results = cls._Phi_lowSiO2(molar_SiO2, molar_Na2O, molar_K2O)
                results[high_SiO2] = cls._Phi_highSiO2(
                    molar_SiO2[high_SiO2], molar_Na2O[high_SiO2], molar_K2O[high_SiO2]
                )
                return results

        except TypeError:
            if molar_SiO2 > 60:
                # raise RuntimeError("SiO2 >60 mol%")
                return cls._Phi_highSiO2(molar_SiO2, molar_Na2O, molar_K2O)

        return cls._Phi_lowSiO2(molar_SiO2, molar_Na2O, molar_K2O)

    @classmethod
    def _Phi_highSiO2(cls, molar_SiO2, molar_Na2O, molar_K2O):
        """
        Phi for SiO2 > 60 mol%
        """

        return (11 - 5.5 * 100 / (100 - molar_SiO2)) * np.exp(
            -0.31 * (molar_Na2O + molar_K2O)
        )

    @classmethod
    def _Phi_lowSiO2(cls, molar_SiO2, molar_Na2O, molar_K2O):
        """
        Phi for SiO2 < 60 mol%
        """

        return (0.46 * (100 / (100 - molar_SiO2)) - 0.93) * (molar_Na2O + molar_K2O) + (
            -5.33 * (100 / (100 - molar_SiO2)) + 9.69
        )

    @classmethod
    def _SiO2_A(cls, melt_mol_fractions):
        """returns adjusted SiO2 for Toplis (2005) Fe-Mg Kd calculations

        Equations 11 and 14 calculate adjusted molar SiO2 by correcting for akalis and water


        Parameters
        ----------
        molar_SiO2 : int or list-like

        molar_Na2O : int or list-like

        molar_K2O : int or list-like

        Phi : int or list-like
            coefficient for alkali correction, needs to be calculated according to Toplis (eq 12, 2005)

        H2O : int or list-like, optional
            wt. %


        Returns
        -------
        int or list-like
        """

        # Calculate melt molar concentrations
        # Molar fractions normalised to 1
        melt_mol_fractions = melt_mol_fractions.fillna(0.0)
        molar_concentrations = melt_mol_fractions * 100

        molar_SiO2 = molar_concentrations["SiO2"]
        molar_Na2O = molar_concentrations["Na2O"]
        molar_K2O = molar_concentrations["K2O"]

        Phi = cls._Phi(molar_SiO2, molar_Na2O, molar_K2O)
        # Equation 11
        SiO2_A = molar_SiO2 + Phi * (molar_Na2O + molar_K2O)

        try:
            # For dataframes
            if "H2O" in molar_concentrations.columns:
                SiO2_A = SiO2_A + 0.8 * molar_concentrations["H2O"]  # equation 14
        except:
            # For series
            if "H2O" in molar_concentrations.index:
                SiO2_A = SiO2_A + 0.8 * molar_concentrations["H2O"]  # equation 14

        return SiO2_A

    @classmethod
    @np.errstate(invalid="raise")
    def _calculate_Kd(
        cls,
        melt_mol_fractions: pd.DataFrame,
        forsterite: float | pd.Series,
        T_K: float | pd.Series,
        P_bar: float | pd.Series,
        *args,
        **kwargs,
    ) -> float | pd.Series:
        """
        Calculate Kds for given melt compositions and fixed forsterite content.

        Parameters
        ----------
        melt_mol_fractions : pandas Dataframe
            melt compositions in oxide mol fractions
        forsterite : float, array-like
            olivine forsterite contents.
        T_K : float, array-like
            temperatures in Kelvin
        P_bar: float, array-like
            pressures in bar

        Returns
        -------
        Kds : array-like
            Fe-Mg partition coefficients
        """

        SiO2_A = cls._SiO2_A(melt_mol_fractions)

        return np.exp(
            (-6766 / (R * T_K) - 7.34 / R)
            + np.log(0.036 * SiO2_A - 0.22)
            + (3000 * (1 - 2 * forsterite) / (R * T_K))
            + (0.035 * (P_bar - 1) / (R * T_K))
        )

    @classmethod
    def calculate_Kd(
        cls,
        melt_mol_fractions: pd.Series | pd.DataFrame,
        Fe3Fe2: float | pd.Series,
        T_K: float | pd.Series,
        P_bar: float | pd.Series,
        forsterite_initial: float | pd.Series = 0.85,
        *args,
        **kwargs,
    ) -> float | pd.Series:
        """
        Calculate Kds for given melt compositions and equilibriutm forsterite content.

        Parameters
        ----------
        melt_mol_fractions : pandas Dataframe
            melt compositions in oxide mol fractions
        forsterite_initial : float, array-like
            initial olivine forsterite contents. Forsterite values are iteratively adjusted until equilibrium with the melt is reached.
        Fe3Fe2 : float, array-like
            melt Fe3+/Fe2+ ratios
        T_K : float, array-like
            temperatures in Kelvin
        P_bar: float, array-like
            pressures in bar

        Returns
        -------
        Kds : array-like
            Fe-Mg partition coefficients
        """

        if isinstance(melt_mol_fractions, pd.Series):
            Kd_func = iterate_Kd_scalar
        elif isinstance(melt_mol_fractions, pd.DataFrame):
            Kd_func = iterate_Kd_vectorized

        return Kd_func(
            melt_mol_fractions=melt_mol_fractions,
            forsterite_initial=forsterite_initial,
            Fe3Fe2=Fe3Fe2,
            T_K=T_K,
            P_bar=P_bar,
            Kd_model=cls._calculate_Kd,
            *args,
            **kwargs,
        )

    @classmethod
    def get_error(cls) -> float:
        """
        Calculate one standard deviation errors on Kds

        Returns
        -------
        error:  float
            one standard deviation error
        """

        return cls.error

    @classmethod
    def get_offset_parameters(cls, n=1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs):
        error = cls.get_error()
        return offset_parameters * error


class blundy(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 8 from Blundy (2020)\ [20]_.
    """

    errors = pd.Series({6: 0.019, 9: 0.04, 100: 0.063})

    @classmethod
    def _get_Fe3FeTotal(cls, Fe3Fe2, **kwargs):

        # Fe3Fe2_model = Fe_redox.borisov
        # dfO2 = kwargs.get("dfO2", configuration.dfO2)

        # fO2 = calculate_fO2(T_K=T_K, P_bar=P_bar, dfO2=dfO2)
        # Fe3Fe2 = Fe3Fe2_model.calculate_Fe3Fe2(
        #     melt_mol_fractions=melt_mol_fractions, T_K=T_K, fO2=fO2, P_bar=P_bar
        # )

        return Fe3Fe2 / (1 + Fe3Fe2)

    @classmethod
    def _calculate_Kd(
        cls, forsterite, T_K, Fe3Fe2, *args, **kwargs
    ) -> float | pd.Series:
        """
        calculate equilibrium Kds for given melt compositions.

        Parameters
        ----------
        forsterite  : float, array-like
            initial olivine forsterite content. Forsterite values are iteratively adjusted and initial values are not necessarily in Fe-Mg equilibrium with melts.
        T_K : float, array-like
            temperatures in Kelvin
        P_bar   : float array-like
            pressures in bar

        Returns
        -------
        Kds : float, array-like
        """

        Fe3FeTotal = cls._get_Fe3FeTotal(Fe3Fe2=Fe3Fe2)

        Kd_Fe_total = (
            0.3642 * (1 - Fe3FeTotal) * np.exp((312.7 * (1 - 2 * forsterite)) / T_K)
        )

        Kd_Fe2 = Kd_Fe_total / (1 - Fe3FeTotal)

        return Kd_Fe2

    @classmethod
    def calculate_Kd(
        cls,
        melt_mol_fractions: pd.Series | pd.DataFrame,
        T_K: float | pd.Series,
        P_bar: float | pd.Series,
        forsterite_initial: float | pd.Series = 0.85,
        *args,
        **kwargs,
    ) -> float | pd.Series:
        """
        Calculate Kds for given melt compositions and equilibriutm forsterite content.

        Parameters
        ----------
        melt_mol_fractions : pandas Dataframe
            melt compositions in oxide mol fractions
        forsterite_initial : float, array-like
            initial olivine forsterite contents. Forsterite values are iteratively adjusted until equilibrium with the melt is reached.
        T_K : float, array-like
            temperatures in Kelvin
        P_bar: float, array-like
            pressures in bar

        Returns
        -------
        Kds : array-like
            Fe-Mg partition coefficients
        """
        # Fe3Fe2 needs to be calculated with 'borisov' and should be removed from kwargs if present.
        _ = kwargs.pop("Fe3Fe2", None)

        if isinstance(melt_mol_fractions, pd.Series):
            Kd_func = iterate_Kd_scalar
        elif isinstance(melt_mol_fractions, pd.DataFrame):
            Kd_func = iterate_Kd_vectorized

        dfO2 = kwargs.get("dfO2", configuration.dfO2)

        fO2 = calculate_fO2(T_K=T_K, P_bar=P_bar, dfO2=dfO2)
        Fe3Fe2 = Fe3Fe2_models.borisov.calculate_Fe3Fe2(
            melt_mol_fractions=melt_mol_fractions, T_K=T_K, fO2=fO2, P_bar=P_bar
        )

        model = partial(cls._calculate_Kd, Fe3Fe2=Fe3Fe2)

        return Kd_func(
            melt_mol_fractions=melt_mol_fractions,
            forsterite_initial=forsterite_initial,
            T_K=T_K,
            P_bar=P_bar,
            Fe3Fe2=Fe3Fe2,
            Kd_model=model,
            *args,
            **kwargs,
        )

    @classmethod
    def get_error(
        cls, melt_composition: pd.DataFrame, *args, **kwargs
    ) -> float | pd.Series:
        """
        Calculate one standard deviation errors on Kds

        Parameters
        ----------
        melt_composition    : pandas Dataframe
            melt composition in oxide wt. %

        Returns
        -------
        error:  float, array-like
            one standard deviation error
        """

        axis = [0, 1][isinstance(melt_composition, pd.DataFrame)]
        alkalis = melt_composition.wt_pc()[["Na2O", "K2O"]].sum(axis=axis)

        if isinstance(alkalis, (int, float)):

            composition_filter = np.less(alkalis, cls.errors.index.values)
            idx = cls.errors.index.values[composition_filter][0]

            return cls.errors[idx]

        errors = pd.Series(index=melt_composition.index)

        for i, a in alkalis.items():
            composition_filter = np.less(a, cls.errors.index.values)
            idx = cls.errors.index.values[composition_filter][0]
            errors.loc[i] = cls.errors[idx]

        return errors

    @classmethod
    def get_offset_parameters(cls, n=1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, melt_composition, offset_parameters, *args, **kwargs):

        error = cls.get_error(melt_composition=melt_composition)

        return offset_parameters * error


class putirka2016_8a(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 8a from Putirka (2016)\ [21]_.
    """

    Kd_error = 4.4e-2

    @staticmethod
    def calculate_Kd(melt_mol_fractions, *args, **kwargs) -> float:
        """
        Calculate mineral-melt partition coefficients

        Returns
        -------
        float
            mineral-melt partition coefficients
        """

        if isinstance(melt_mol_fractions, (pd.Series, np.ndarray)):
            return 0.33
        elif isinstance(melt_mol_fractions, pd.DataFrame):
            return pd.Series(0.33, index=melt_mol_fractions.index)

    @classmethod
    def get_error(cls, *args, **kwargs) -> float:
        """
        Return one standard deviation errors on partition coefficients.

        Returns
        -------
        float, array-like
            partition coefficient errors
        """
        return cls.Kd_error

    @staticmethod
    def get_offset_parameters(n: int, *args, **kwargs) -> float | np.ndarray:
        """
        Randomly sample a standard normal distribution *n* times.

        n   : int
            sample amount.
        """
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs) -> float | np.ndarray:
        """
        Calculate random samples of partition coefficient errors

        Parameters
        ----------
        offset_parameters : float, array-like
            random samples of a standard normal distribution.
        """
        return cls.get_error() * offset_parameters


class putirka2016_8b(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 8b from Putirka (2016)\ [21]_.

    For P > 1 GPa
    """

    a = 0.21
    b = 8e-3
    c = 2.5e-3
    d = -3.63e-4

    components = ["SiO2", "Na2O", "K2O"]

    Kd_error = 4.4e-2

    @classmethod
    def calculate_Kd(
        cls, melt_mol_fractions, P_bar, *args, **kwargs
    ) -> float | np.ndarray:
        """
        Calculate mineral-melt partition coefficients

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        P_bar :   float, array-like
            pressure in bar

        Returns
        -------
        float, array-like
            mineral-melt partition coefficients
        """
        wt_pc = melt_mol_fractions.wt_pc()
        wt_pc = check_components(wt_pc, cls.components)
        P_GPa = P_bar / 1e4
        axis = [0, 1][isinstance(wt_pc, pd.DataFrame)]

        return (
            cls.a
            + cls.b * P_GPa
            + cls.c * wt_pc["SiO2"]
            + cls.d * (wt_pc[["Na2O", "K2O"]].sum(axis=axis) ** 2)
        )

    @classmethod
    def get_error(cls, *args, **kwargs) -> float:
        """
        Return one standard deviation errors on partition coefficients.

        Returns
        -------
        float, array-like
            partition coefficient errors
        """
        return cls.Kd_error

    @staticmethod
    def get_offset_parameters(n: int, *args, **kwargs) -> float | np.ndarray:
        """
        Randomly sample a standard normal distribution *n* times.

        n   : int
            sample amount.
        """
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs) -> float | np.ndarray:
        """
        Calculate random samples of partition coefficient errors

        Parameters
        ----------
        offset_parameters : float, array-like
            random samples of a standard normal distribution.
        """
        return cls.get_error() * offset_parameters


class putirka2016_8c(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 8c from Putirka (2016)\ [21]_.

    for P < 1 GPa
    """

    a = 0.25
    b = 1.8e-3
    c = -3.25e-4

    components = ["SiO2", "Na2O", "K2O"]

    Kd_error = 4e-2

    @classmethod
    def calculate_Kd(cls, melt_mol_fractions, *args, **kwargs) -> float | np.ndarray:
        """
        Calculate mineral-melt partition coefficients

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions


        Returns
        -------
        float, array-like
            mineral-melt partition coefficients
        """
        wt_pc = melt_mol_fractions.wt_pc()
        wt_pc = check_components(wt_pc, cls.components)
        axis = [0, 1][isinstance(wt_pc, pd.DataFrame)]

        return (
            cls.a
            + cls.b * wt_pc["SiO2"]
            + cls.c * (wt_pc[["Na2O", "K2O"]].sum(axis=axis) ** 2.0)
        )

    @classmethod
    def get_error(cls, *args, **kwargs) -> float:
        """
        Return one standard deviation errors on partition coefficients.

        Returns
        -------
        float, array-like
            partition coefficient errors
        """
        return cls.Kd_error

    @staticmethod
    def get_offset_parameters(n: int, *args, **kwargs) -> float | np.ndarray:
        """
        Randomly sample a standard normal distribution *n* times.

        n   : int
            sample amount.
        """
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs) -> float | np.ndarray:
        """
        Calculate random samples of partition coefficient errors

        Parameters
        ----------
        offset_parameters : float, array-like
            random samples of a standard normal distribution.
        """
        return cls.get_error() * offset_parameters


class putirka2016_8d(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 8d from Putirka (2016)\ [21]_.

    For liquid compositions with <45 wt.% SiO2 and > 8 wt.% Na2O + K2O
    """

    a = 0.6
    b = 1.3e-2
    c = 1.6e-2
    d = -1.73e-4
    e = 1.79e-2
    f = -2.6
    g = 2.11e-1
    h = 3.19e-5

    components = ["SiO2", "Al2O3", "Na2O", "K2O"]

    components = ["SiO2", "Na2O", "K2O"]

    Kd_error = 4.2e-2

    @classmethod
    def calculate_Kd(
        cls, melt_mol_fractions, T_K, P_bar, *args, **kwargs
    ) -> float | np.ndarray:
        """
        Calculate mineral-melt partition coefficients

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        P_bar :   float, array-like
            pressure in bar

        Returns
        -------
        float, array-like
            mineral-melt partition coefficients
        """
        wt_pc = melt_mol_fractions.wt_pc()
        wt_pc = check_components(wt_pc, components=cls.components)

        P_GPa = P_bar / 1e4

        axis = [0, 1][isinstance(wt_pc, pd.DataFrame)]

        Al_number = wt_pc["Al2O3"] / wt_pc[["Al2O3", "SiO2"]].sum(axis=axis)

        return (
            cls.a
            + cls.b * P_GPa
            + cls.c * wt_pc["SiO2"]
            + cls.d * (wt_pc["SiO2"] ** 2.0)
            + cls.e * wt_pc["Al2O3"]
            + cls.f * Al_number
            + cls.g * np.log(Al_number)
            + cls.h * (wt_pc[["Na2O", "K2O"]].sum(axis=axis) ** 3.0)
        )

    @classmethod
    def get_error(cls, *args, **kwargs) -> float:
        """
        Return one standard deviation errors on partition coefficients.

        Returns
        -------
        float, array-like
            partition coefficient errors
        """
        return cls.Kd_error

    @staticmethod
    def get_offset_parameters(n: int, *args, **kwargs) -> float | np.ndarray:
        """
        Randomly sample a standard normal distribution *n* times.

        n   : int
            sample amount.
        """
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs) -> float | np.ndarray:
        """
        Calculate random samples of partition coefficient errors

        Parameters
        ----------
        offset_parameters : float, array-like
            random samples of a standard normal distribution.
        """
        return cls.get_error() * offset_parameters


class sun2020(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 7 from Sun & Dasgupta (2020)\ [22]_.

    Sun, C., Dasgupta, R. (2020) Thermobarometry of CO2-rich, silica-undersaturated melts constrains cratonic lithosphere thinning through time in areas of kimberlitic magmatism. Earth and Planetary Sience Letters. 550
    """

    volatiles = ["H2O", "CO2", "F", "S", "Cl"]
    components = ["MgO", "Na2O", "H2O"]
    error = 0.03

    @classmethod
    @np.errstate(invalid="raise")
    def calculate_Kd(
        cls, melt_mol_fractions, Fe3Fe2, *args, **kwargs
    ) -> float | pd.Series:
        """
        Calculate Kds for given melt compositions and fixed forsterite content.

        Parameters
        ----------
        melt_mol_fractions : pandas Dataframe
            melt compositions in oxide mol fractions
        forsterite : float, array-like
            olivine forsterite contents.
        T_K : float, array-like
            temperatures in Kelvin
        P_bar: float, array-like
            pressures in bar

        Returns
        -------
        Kds : array-like
            Fe-Mg partition coefficients
        """
        composition = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        melt_volatile_free = _remove_elements(composition, drop=cls.volatiles)
        melt_per_oxygen = moles_per_oxygen(moles=melt_volatile_free)
        melt_wtpc = composition.wt_pc()

        Kd_Fe_total = np.exp(
            -1.65
            + 1.22 * np.sqrt(melt_per_oxygen["Mg1O"])
            + 2.45 * melt_per_oxygen["Na2O"]
            + 0.54 * melt_wtpc["H2O"] / 100
        )
        Fe3FeTotal = Fe3Fe2 / (1 + Fe3Fe2)

        Kd_Fe2 = Kd_Fe_total / (1 - Fe3FeTotal)

        return Kd_Fe2

    @classmethod
    def get_error(cls) -> float:
        """
        Calculate one standard deviation errors on Kds

        Returns
        -------
        error:  float
            one standard deviation error
        """

        return cls.error

    @classmethod
    def get_offset_parameters(cls, n=1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs):
        error = cls.get_error()
        return offset_parameters * error


_clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
# Collect all Kd_models in a dictionary.
Kd_olmelt_FeMg_models_dict = {
    cls[0]: cls[1] for cls in _clsmembers if _is_Kd_model(cls[1])
}



from abc import ABC, abstractmethod

import numpy as np


class Kd_model(ABC):
    @abstractmethod
    def calculate_Kd(
        cls, melt_mol_fractions, T_K, P_bar, *args, **kwargs
    ) -> float | np.ndarray:
        """
        Calculate mineral-melt partition coefficients

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        P_bar :   float, array-like
            pressure in bar

        Returns
        -------
        float, array-like
            mineral-melt partition coefficients
        """
        pass

    @classmethod
    def _calculate_Kd_(
        cls, melt_mol_fractions, T_K, P_bar, offset_parameters=0.0, *args, **kwargs
    ):

        Kd = cls.calculate_Kd(
            melt_mol_fractions=melt_mol_fractions, T_K=T_K, P_bar=P_bar, *args, **kwargs
        )

        if offset_parameters == 0.0:
            return Kd

        offset = cls.get_offset(
            melt_composition=melt_mol_fractions,
            offset_parameters=offset_parameters,
        )
        Kd = Kd + offset

        # Make sure no negative values are returned
        try:
            Kd[Kd <= 0.0] = 1e-6
        except TypeError:
            Kd = 1e-6 if Kd <= 0.0 else Kd

        # Make sure arrays of length 1 are returned as floats instead
        try:
            if len(Kd) == 1:
                return np.array(Kd)[0]
            return Kd

        except TypeError:
            return Kd

    @abstractmethod
    def get_error(cls, *args, **kwargs):
        """
        Return one standard deviation errors on partition coefficients.

        Returns
        -------
        float, array-like
            partition coefficient errors
        """
        pass

    @abstractmethod
    def get_offset_parameters(cls, n: int, *args, **kwargs):
        """
        Randomly sample a standard normal distribution *n* times.

        n   : int
            sample amount.
        """
        pass

    @abstractmethod
    def get_offset(cls, offset_parameters, *args, **kwargs):
        """
        Calculate random samples of partition coefficient errors

        Parameters
        ----------
        offset_parameters : float, array-like
            random samples of a standard normal distribution.
        """
        pass



from typing import Union

import numpy as np
import pandas as pd

from MagmaPandas.parse_io import check_components

"""
Parameterisation by:

Giordano, D., Russel, J. & Dingwell, D. (2008) Viscosity of magmatic liquids a model. Earth and planetary science letters. Vol 271, issue 1-4, pp. 123-135.
"""

parameters = pd.Series(
    {
        "A": -4.55,
        "b1": 159.6,
        "b2": -173.3,
        "b3": 72.1,
        "b4": 75.7,
        "b5": -39.0,
        "b6": -84.1,
        "b7": 141.5,
        "b11": -2.43,
        "b12": -0.91,
        "b13": 17.6,
        "c1": 2.75,
        "c2": 15.7,
        "c3": 8.3,
        "c4": 10.2,
        "c5": -12.3,
        "c6": -99.5,
        "c11": 0.3,
    }
)

components = [
    "SiO2",
    "TiO2",
    "Al2O3",
    "FeO",
    "MnO",
    "P2O5",
    "MgO",
    "CaO",
    "Na2O",
    "H2O",
    "F2",
    "K2O",
]


def _calculate_B(melt_mol_fractions: Union[pd.DataFrame, pd.Series]) -> pd.Series:

    axis = [0, 1][isinstance(melt_mol_fractions, pd.DataFrame)]

    mol_percent = check_components(melt_mol_fractions, components=components) * 100

    part_1 = (
        parameters["b1"] * mol_percent[["SiO2", "TiO2"]].sum(axis=axis)
        + parameters["b2"] * mol_percent["Al2O3"]
        + parameters["b3"] * mol_percent[["FeO", "MnO", "P2O5"]].sum(axis=axis)
        + parameters["b4"] * mol_percent["MgO"]
        + parameters["b5"] * mol_percent["CaO"]
        + parameters["b6"] * mol_percent[["Na2O", "H2O", "F2"]].sum(axis=axis)
        + parameters["b7"]
        * (mol_percent[["H2O", "F2"]].sum(axis=axis) + np.log(1 + mol_percent["H2O"]))
    )

    part_2 = (
        parameters["b11"]
        * (
            mol_percent[["SiO2", "TiO2"]].sum(axis=axis)
            * mol_percent[["FeO", "MnO", "MgO"]].sum(axis=axis)
        )
        + parameters["b12"]
        * (
            mol_percent[["SiO2", "TiO2", "Al2O3", "P2O5"]].sum(axis=axis)
            * mol_percent[["Na2O", "K2O", "H2O"]].sum(axis=axis)
        )
        + parameters["b13"]
        * (mol_percent["Al2O3"] * mol_percent[["Na2O", "K2O"]].sum(axis=axis))
    )

    return part_1 + part_2


def _calculate_C(melt_mol_fractions: Union[pd.DataFrame, pd.Series]) -> pd.Series:

    axis = [0, 1][isinstance(melt_mol_fractions, pd.DataFrame)]

    mol_percent = check_components(melt_mol_fractions, components=components) * 100

    part_1 = (
        parameters["c1"] * mol_percent["SiO2"]
        + parameters["c2"] * mol_percent[["TiO2", "Al2O3"]].sum(axis=axis)
        + parameters["c3"] * mol_percent[["FeO", "MgO", "MnO"]].sum(axis=axis)
        + parameters["c4"] * mol_percent["CaO"]
        + parameters["c5"] * mol_percent[["Na2O", "K2O"]].sum(axis=axis)
        + parameters["c6"] * np.log(1 + mol_percent[["H2O", "F2"]].sum(axis=axis))
    )

    part_2 = parameters["c11"] * (
        (
            mol_percent[["Al2O3", "FeO", "MgO", "MnO", "CaO"]].sum(axis=axis)
            - mol_percent["P2O5"]
        )
        * mol_percent[["Na2O", "K2O", "H2O", "F2"]].sum(axis=axis)
    )

    return part_1 + part_2


def calculate_viscosity(
    melt_mol_fractions: Union[pd.DataFrame, pd.Series], T_K: Union[pd.Series, float]
) -> pd.Series:
    """
    Calculate melt viscosity according to equation 1 of Giordano et al. (2008)\ [27]_.

    Parameters
    ----------
    melt_mol_fractions
        melt mol fractions
    T_K
        temperature in Kelvin

    Returns
    -------
        viscosity in log10(Pa.s)
    """

    B = _calculate_B(melt_mol_fractions=melt_mol_fractions)
    C = _calculate_C(melt_mol_fractions=melt_mol_fractions)

    return parameters["A"] + B / (T_K - C)



"""
Module with |CO2|-|H2O| solublity models for silicate melts
"""

from MagmaPandas.volatile_solubility.calculate_volatiles import *
from MagmaPandas.volatile_solubility.isobars import *
from MagmaPandas.volatile_solubility.solubility_baseclass import Solubility_model



from importlib import resources

import pandas as pd

REE_radii = pd.Series(
    {
        "La": 1.16,
        "Ce": 1.143,
        "Pr": 1.126,
        "Nd": 1.109,
        "Sm": 1.079,
        "Eu": 1.066,
        "Gd": 1.053,
        "Tb": 1.040,
        "Dy": 1.027,
        "Y": 1.019,
        "Ho": 1.015,
        "Er": 1.004,
        "Tm": 0.994,
        "Yb": 0.985,
        "Lu": 0.977,
    }
)

REE = REE_radii.index.values

divalent_radii = pd.Series({"Mg": 0.89, "Ba": 1.42, "Ca": 1.12, "Eu": 1.25, "Sr": 1.26})


def C1_chondrite() -> pd.Series:
    """Returns C1 chondrite composition from McDonough & Sun (1995)"""

    with resources.open_text("geoplot.data", "Mcdonough_sun_1995.csv") as df:
        C1 = pd.read_csv(df, index_col=["reservoir"]).loc["C1"]

    return C1


def primitive_mantle() -> pd.Series:
    """Returns primitive mantle composition from McDonough & Sun (1995)"""

    with resources.open_text("geoplot.data", "Mcdonough_sun_1995.csv") as df:
        PM = pd.read_csv(df, index_col=["reservoir"]).loc["Pyrolite"]

    return PM


def MORB(type: str, add_majors=False, conf_interval=False) -> pd.Series:
    """Returns primitive mantle composition from McDonough & Sun (1995)"""

    with resources.open_text("geoplot.data", "MORB_Gale2013.csv") as df:
        MORB = pd.read_csv(df, index_col=["reservoir"])

    major_elements = [
        "MgO",
        "SiO2",
        "FeO",
        "CaO",
        "Na2O",
        "Al2O3",
        "TiO2",
        "K2O",
        "P2O5",
        "MnO",
    ]

    if not add_majors:
        MORB = MORB.drop(columns=major_elements)

    if not conf_interval:
        return MORB.loc[type]

    confidence95 = MORB.loc[f"{type}_conf95"]

    return MORB.loc[type], confidence95


def radii(valency):
    """Returns a dictionary with radii for common trace elements

    Parameters
    ----------
    valency : {2, 3, 'REE'}
        valency of requisted trace elements

    Returns
    -------
    dictionary
        dictionary with ionic radii in Angstrom
    """

    if valency not in [2, 3, "REE"]:
        raise Exception("valency should be 2, 3 or 'REE'")

    valency = {2: 0, 3: 1, "REE": 1}[valency]

    divalent = pd.Series({"Mg": 0.89, "Ba": 1.42, "Ca": 1.12, "Eu": 1.25, "Sr": 1.26})

    REE = pd.Series(
        {
            "La": 1.16,
            "Ce": 1.143,
            "Pr": 1.126,
            "Nd": 1.109,
            "Sm": 1.079,
            "Eu": 1.066,
            "Gd": 1.053,
            "Tb": 1.040,
            "Dy": 1.027,
            "Y": 1.019,
            "Ho": 1.015,
            "Er": 1.004,
            "Tm": 0.994,
            "Yb": 0.985,
            "Lu": 0.977,
        }
    )

    total = [divalent, REE]

    return total[valency]
